Type: Package
Package: MazamaCoreUtils
Version: 0.6.2
Title: Utility Functions for Production R Code
Authors@R: c(
    person("Jonathan", "Callahan", email="jonathan.s.callahan@gmail.com", role=c("aut","cre")),
    person("Eli", "Grosman", email="eligrosman1@gmail.com", role=c("ctb")),
    person("Spencer", "Pease", email="spencerpease618@gmail.com", role=c("ctb")),
    person("Thomas", "Bergamaschi", email="tbergama@gmail.com", role=c("ctb"))
    )
Maintainer: Jonathan Callahan <jonathan.s.callahan@gmail.com>
Description: A suite of utility functions providing functionality commonly
    needed for production level projects such as logging, error handling,
    cache management and date-time parsing. Functions for date-time parsing and 
    formatting require that time zones be specified explicitly, avoiding a common 
    source of error when working with environmental time series.
License: GPL-3
URL: https://github.com/MazamaScience/MazamaCoreUtils
BugReports: https://github.com/MazamaScience/MazamaCoreUtils/issues
Depends: R (>= 4.0.0)
Imports: devtools, dplyr, geohashTools, logger, lubridate, magrittr,
        purrr, rlang (>= 1.1.0), rvest, stringr, tibble, xml2
Suggests: knitr, markdown, testthat (>= 3.1.7), rmarkdown, roxygen2
Encoding: UTF-8
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-05 21:24:15 UTC; jonathancallahan
Author: Jonathan Callahan [aut, cre],
  Eli Grosman [ctb],
  Spencer Pease [ctb],
  Thomas Bergamaschi [ctb]
Repository: CRAN
Date/Publication: 2026-05-06 05:10:36 UTC
Built: R 4.6.0; ; 2026-05-06 07:08:22 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(DEBUG)
export(ERROR)
export(FATAL)
export(INFO)
export(TRACE)
export(WARN)
export(check)
export(check_fast)
export(check_faster)
export(check_fastest)
export(check_slow)
export(check_slower)
export(check_slowest)
export(createLocationID)
export(createLocationMask)
export(dateRange)
export(dateSequence)
export(getAPIKey)
export(html_getLinkNames)
export(html_getLinkUrls)
export(html_getLinks)
export(html_getTable)
export(html_getTables)
export(initializeLogging)
export(lintFunctionArgs_dir)
export(lintFunctionArgs_file)
export(loadDataFile)
export(logger.debug)
export(logger.error)
export(logger.fatal)
export(logger.info)
export(logger.isInitialized)
export(logger.setLevel)
export(logger.setup)
export(logger.trace)
export(logger.warn)
export(manageCache)
export(parseDatetime)
export(setAPIKey)
export(setIfNull)
export(showAPIKeys)
export(stopIfNull)
export(stopOnError)
export(timeRange)
export(timeStamp)
export(timezoneLintRules)
export(validateLonLat)
export(validateLonsLats)
importFrom(logger,appender_console)
importFrom(logger,appender_file)
importFrom(logger,appender_tee)
importFrom(logger,appender_void)
importFrom(logger,formatter_sprintf)
importFrom(logger,layout_simple)
importFrom(logger,log_appender)
importFrom(logger,log_debug)
importFrom(logger,log_error)
importFrom(logger,log_fatal)
importFrom(logger,log_formatter)
importFrom(logger,log_info)
importFrom(logger,log_layout)
importFrom(logger,log_threshold)
importFrom(logger,log_trace)
importFrom(logger,log_warn)
importFrom(magrittr,"%>%")
importFrom(rlang,.data)
importFrom(utils,str)

[![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/MazamaCoreUtils)](https://cran.r-project.org/package=MazamaCoreUtils)
[![Downloads](https://cranlogs.r-pkg.org/badges/MazamaCoreUtils)](https://cran.r-project.org/package=MazamaCoreUtils)
[![DOI](https://zenodo.org/badge/152321630.svg)](https://zenodo.org/badge/latestdoi/152321630)

A dedicated Slack channel has been created for announcements, support and to 
help build a community of practice around this open source package. You may 
request an invitation to join from <jonathan.s.callahan@gmail.com>.

# MazamaCoreUtils

```
A suite of utility functions providing functionality commonly needed for 
production level projects such as logging, error handling, cache management and 
date-time parsing. Functions for date-time parsing and formatting require that 
timezones be specified explicitly, avoiding a common source of error when 
working with environmental time series.
```

## Background

The MazamaCoreUtils package was created by Mazama Science to regularize
work building R packages, data processing pipelines and web services focused on 
environmental monitoring data.

The main goal of this package is to create an internally standardized set of
functions that for use in various systems that are being run operationally. 
Areas of functionality supported by this package include:

 * python style logging
 * simple error messaging
 * cache management
 * API key handling
 * date-time parsing and formatting
 * lat/lon validation and uniqueID creation
 * source code linting
 
## Installation

Install from CRAN with:

```install.packages('MazamaCoreUtils')```

Install the latest version from GitHub with:

```devtools::install_github('mazamascience/MazamaCoreUtils')```

----

Development of this R package has been supported with funding from the 
following institutions:

* USFS [AirFire Research Team](https://www.airfire.org)

Questions regarding further development of the package should be directed to 
<jonathan.s.callahan@gmail.com>.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----stopOnError--------------------------------------------------------------
library(MazamaCoreUtils)
logger.setup()
logger.setLevel(TRACE) # force logs to be printed to the console

# Arbitrarily deep in the stack we might have:
myFunc <- function(x) {
  return(log(x))
}

# ----- Example 1:  good user input --------------------------------------------
try({
  
  userInput <- 10
  logger.trace("class(userInput) = %s", class(userInput))
  
  try({
    myFunc(x = userInput)
  }, silent = TRUE) %>%
    stopOnError()
  
  logger.trace("Continue processing ...")
  
}, silent = TRUE)

# ----- Example 2:  bad user input ---------------------------------------------
try({
  
  userInput <- "10"
  logger.trace("class(userInput) = %s", class(userInput))
  
  try({
    myFunc(x = userInput)
  }, silent = TRUE) %>%
    stopOnError()
  
  logger.trace("Continue processing ...") # we don't get here
  
}, silent = TRUE)

# ----- Example 3:  bad user input, custom error message -----------------------
try({
  
  try({
    logger.trace("class(userInput) = %s", class(userInput))
    myFunc(x = userInput)
  }, silent = TRUE) %>%
    stopOnError("Unable to process user input")
  
  logger.trace("Continue processing ...") # we don't get here
  
}, silent = TRUE)

test_that("stopOnError() is silent when there is no error", {

  # setup
  myFunc <- function(x) a <- log(x)

  result <- try(myFunc(10), silent = TRUE)

  expect_silent(stopOnError(result))
  expect_silent(stopOnError(result, "custom error message"))

})

test_that("stopOnError() converts error messages", {

  # setup
  myFunc <- function(x) a <- log(x)

  result <- try(myFunc("ten"), silent = TRUE)

  expect_error(stopOnError(result))
  expect_error(stopOnError(result, "custom error message"), "custom error message")

})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----createCache--------------------------------------------------------------
# Create a cache directory
CACHE_DIR <- file.path(tempdir(), 'cache')
if ( file.exists(CACHE_DIR) == FALSE ) {
  dir.create(CACHE_DIR)
}

# Add a few files to the cache
write.csv(matrix(1,400,500), file=file.path(CACHE_DIR,'m1.csv'))
Sys.sleep(1) # wait a bit between each to give them different mtimes
write.csv(matrix(2,400,500), file=file.path(CACHE_DIR,'m2.csv'))
Sys.sleep(1)
write.csv(matrix(3,400,500), file=file.path(CACHE_DIR,'m3.csv'))
Sys.sleep(1)
write.csv(matrix(4,400,500), file=file.path(CACHE_DIR,'m4.csv'))

## ----checkCache---------------------------------------------------------------
cachedFiles <- list.files(CACHE_DIR, full.names = TRUE)
infoDF <- file.info(cachedFiles)
cacheSize = (sum(infoDF$size) / 1e6) # in MB
print(list.files(CACHE_DIR))
sprintf("Cache size = %s MB", cacheSize)

## ----accessFiles, echo=TRUE---------------------------------------------------
# Access two of the files, updating their atime
invisible( read.csv(file.path(CACHE_DIR, 'm1.csv')) )
invisible( read.csv(file.path(CACHE_DIR, 'm2.csv')) )

## ----manageCache--------------------------------------------------------------
# Use manageCache() to get cache to 1 MB
library(MazamaCoreUtils)
manageCache(CACHE_DIR, extensions = 'csv', maxCacheSize = 1)

## ----checkCacheAgain----------------------------------------------------------
# Check cache contents and total size again
cachedFiles <- list.files(CACHE_DIR, full.names = TRUE)
infoDF <- file.info(cachedFiles)
cacheSize = (sum(infoDF$size) / 1e6) # in MB
print(list.files(CACHE_DIR))
sprintf("Cache size = %s MB", cacheSize)

## ----maxFileAge---------------------------------------------------------------
manageCache(CACHE_DIR, maxFileAge = 3/24)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----loggingExample-----------------------------------------------------------
library(MazamaCoreUtils)

# Set up logging 
tmpDir <- tempdir()
logger.setup(
  errorLog = file.path(tmpDir,"ERROR.log"),
  infoLog = file.path(tmpDir,"INFO.log"),
  traceLog = file.path(tmpDir,"TRACE.log")
)

# Begin main program
logger.info("Begin loggingExample ...")

# Try something that might fail
outputSums <- list()
for ( column in names(iris) ) {
  
  logger.trace("Working on: %s", column)
  
  result <- try({
    outputSums[[column]] <- 
      paste0("Sum of ", column, " = ", sum(iris[[column]]))
  }, silent = TRUE)
  
  if ( "try-error" %in% class(result) ) {
    logger.warn("Oops: %s", geterrmessage())
  }
  
}

# Successful completion
logger.error("NO ERRORS")
logger.info("Successfully completed loggingExample")
logger.info("\n=============== THE END ===============\n")

# Check the output
str(outputSums)

## ----errorLog, echo = TRUE----------------------------------------------------
cat(paste(readLines(file.path(tmpDir,"ERROR.log")),collapse='\n'))

## ----infoLog, echo = TRUE-----------------------------------------------------
cat(paste(readLines(file.path(tmpDir,"INFO.log")),collapse='\n'))

## ----traceLog, echo = TRUE----------------------------------------------------
cat(paste(readLines(file.path(tmpDir,"TRACE.log")),collapse='\n'))

#' Create character timestamps
#'
#' Convert datetimes to compact character timestamps suitable for file names,
#' identifiers, labels, and other reproducible text output.
#'
#' Input values are converted with [parseDatetime()] using the required
#' `timezone` argument. When `datetime = NULL`, the current UTC time is used
#' and `timezone` defaults to `"UTC"`.
#'
#' The `unit` argument controls the precision of the output timestamp. The
#' `style` argument controls the output format.
#'
#' Supported `unit` values are:
#'
#' \preformatted{
#' "year"
#' "month"
#' "day"
#' "hour"
#' "min"
#' "sec"
#' "msec"
#' }
#'
#' Supported `style` values are:
#'
#' \preformatted{
#' "ymdhms"   compact calendar time
#' "ymdThms"  compact calendar time with "T" separator
#' "julian"   year and Julian day
#' "clock"    ISO-like clock time
#' }
#'
#' For `style = "julian"` and `unit = "month"`, the timestamp uses the Julian
#' day associated with the beginning of the month.
#'
#' @param datetime Vector of character, integer, or `POSIXct` datetimes.
#' @param timezone Olson timezone used to interpret incoming datetimes.
#' @param unit Temporal precision of the generated timestamp.
#' @param style Output timestamp style.
#'
#' @return
#' Character vector of timestamps.
#'
#' @inheritSection dateRange POSIXct inputs
#'
#' @examples
#' datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")
#'
#' timeStamp()
#' timeStamp(datetime, "UTC", unit = "year")
#' timeStamp(datetime, "UTC", unit = "month")
#' timeStamp(datetime, "UTC", unit = "month", style = "julian")
#' timeStamp(datetime, "UTC", unit = "day")
#' timeStamp(datetime, "UTC", unit = "day", style = "julian")
#' timeStamp(datetime, "UTC", unit = "hour")
#' timeStamp(datetime, "UTC", unit = "min")
#' timeStamp(datetime, "UTC", unit = "sec")
#' timeStamp(datetime, "UTC", unit = "sec", style = "ymdThms")
#' timeStamp(datetime, "UTC", unit = "sec", style = "julian")
#' timeStamp(datetime, "UTC", unit = "sec", style = "clock")
#' timeStamp(datetime, "America/Los_Angeles", unit = "sec", style = "clock")
#' timeStamp(datetime, "America/Los_Angeles", unit = "msec", style = "clock")
#'
#' @export
#'
timeStamp <- function(
  datetime = NULL,
  timezone = NULL,
  unit = "sec",
  style = "ymdhms"
) {

  # ----- Validate parameters --------------------------------------------------

  # Common use case
  if ( is.null(datetime) ) {
    datetime <- lubridate::now(tzone = "UTC")
    if ( is.null(timezone) ) timezone <- "UTC"
  }

  stopIfNull(datetime)
  stopIfNull(timezone)
  stopIfNull(unit)
  stopIfNull(style)

  if ( !timezone %in% base::OlsonNames() )
    stop(sprintf("'timezone = %s' is not found in OlsonNames()", timezone))

  if ( !unit %in% c("year", "month", "day", "hour", "min", "sec", "msec") )
    stop(sprintf("'unit = %s' is not recognized", unit))

  if ( !style %in% c("ymdhms", "ymdThms", "julian", "clock") )
    stop(sprintf("'style = %s' is not recognized", style))

  # ----- Format datetimes -----------------------------------------------------

  # Guarantee conversion to POSIXct
  datetime <- parseDatetime(datetime, timezone = timezone)

  if ( unit == "year" ) {
    format <- "%Y"

  } else if ( unit == "month" ) {
    if ( style == "ymdhms" )
      format <- "%Y%m"
    if ( style == "ymdThms" )
      format <- "%Y%m"
    if ( style == "julian" ) {
      datetime <- lubridate::floor_date(datetime, unit = "month")
      format <- "%Y%j"
    }
    if ( style == "clock" )
      format <- "%Y-%m"

  } else if ( unit == "day" ) {
    if ( style == "ymdhms" )
      format <- "%Y%m%d"
    if ( style == "ymdThms" )
      format <- "%Y%m%d"
    if ( style == "julian" )
      format <- "%Y%j"
    if ( style == "clock" )
      format <- "%Y-%m-%d"

  } else if ( unit == "hour" ) {
    if ( style == "ymdhms" )
      format <- "%Y%m%d%H"
    if ( style == "ymdThms" )
      format <- "%Y%m%dT%H"
    if ( style == "julian" )
      format <- "%Y%j%H"
    if ( style == "clock" )
      format <- "%Y-%m-%dT%H"

  } else if ( unit == "min" ) {
    if ( style == "ymdhms" )
      format <- "%Y%m%d%H%M"
    if ( style == "ymdThms" )
      format <- "%Y%m%dT%H%M"
    if ( style == "julian" )
      format <- "%Y%j%H%M"
    if ( style == "clock" )
      format <- "%Y-%m-%dT%H:%M"

  } else if ( unit == "sec" ) {
    if ( style == "ymdhms" )
      format <- "%Y%m%d%H%M%S"
    if ( style == "ymdThms" )
      format <- "%Y%m%dT%H%M%S"
    if ( style == "julian" )
      format <- "%Y%j%H%M%S"
    if ( style == "clock" )
      format <- "%Y-%m-%dT%H:%M:%S"

  } else if ( unit == "msec" ) {
    if ( style == "ymdhms" )
      format <- "%Y%m%d%H%M%OS3"
    if ( style == "ymdThms" )
      format <- "%Y%m%dT%H%M%OS3"
    if ( style == "julian" )
      format <- "%Y%j%H%M%OS3"
    if ( style == "clock" )
      format <- "%Y-%m-%dT%H:%M:%OS3"

  }

  timeStamp <- strftime(datetime, format = format, tz = timezone)

  # ----- Return ---------------------------------------------------------------

  return(timeStamp)

}

test_that("explicit arguments are required", {

  expect_error(dateSequence(unit = NULL))
  expect_error(dateSequence(style = NULL))

})

test_that("works will NULL datetime", {

  datetime <- lubridate::now(tzone = "UTC")
  stamp <- timeStamp(datetime, "UTC", unit = "hour")
  expect_identical(timeStamp(unit = "hour"), stamp)
  stamp <- timeStamp(datetime, "UTC", unit = "hour", style = "clock")
  expect_identical(timeStamp(unit = "hour", style = "clock"), stamp)

})

test_that("'ymdhms' output is correct", {

  datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")

  expect_identical(timeStamp(datetime, "UTC", unit = "year"), "2019")
  expect_identical(timeStamp(datetime, "UTC", unit = "month"), "201901")
  expect_identical(timeStamp(datetime, "UTC", unit = "day"), "20190108")
  expect_identical(timeStamp(datetime, "UTC", unit = "hour"), "2019010812")
  expect_identical(timeStamp(datetime, "UTC", unit = "min"), "201901081230")
  expect_identical(timeStamp(datetime, "UTC", unit = "sec"), "20190108123015")
  expect_identical(timeStamp(datetime, "UTC", unit = "msec"), "20190108123015.000")

})

test_that("'ymdThms' output is correct", {

  datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")

  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "year"), "2019")
  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "month"), "201901")
  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "day"), "20190108")
  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "hour"), "20190108T12")
  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "min"), "20190108T1230")
  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "sec"), "20190108T123015")
  expect_identical(timeStamp(datetime, "UTC", style = "ymdThms", unit = "msec"), "20190108T123015.000")

})

test_that("'julian' output is correct", {

  datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")

  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "year"), "2019")
  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "month"), "2019001")
  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "day"), "2019008")
  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "hour"), "201900812")
  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "min"), "20190081230")
  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "sec"), "2019008123015")
  expect_identical(timeStamp(datetime, "UTC", style = "julian", unit = "msec"), "2019008123015.000")

})

test_that("'clock' output is correct", {

  datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")

  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "year"), "2019")
  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "month"), "2019-01")
  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "day"), "2019-01-08")
  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "hour"), "2019-01-08T12")
  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "min"), "2019-01-08T12:30")
  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "sec"), "2019-01-08T12:30:15")
  expect_identical(timeStamp(datetime, "UTC", style = "clock", unit = "msec"), "2019-01-08T12:30:15.000")

})

test_that("timezones are respected", {

  # PST
  datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")
  expect_identical(
    timeStamp(datetime, "America/Los_Angeles", style = "clock", unit = "sec"),
    "2019-01-08T04:30:15"
  )

  # PDT
  datetime <- parseDatetime("2019-06-08 12:30:15", timezone = "UTC")
  expect_identical(
    timeStamp(datetime, "America/Los_Angeles", style = "clock", unit = "sec"),
    "2019-06-08T05:30:15"
  )

})

#' Parse datetimes
#'
#' Convert character, numeric, integer, or `POSIXct` datetimes to `POSIXct`.
#'
#' This function accepts a variety of compact date/time formats commonly used in
#' Mazama Science packages, including `Y`, `Ym`, `Ymd`, `YmdH`, `YmdHM`, and
#' `YmdHMS`. Inputs may be mixed within the same vector.
#'
#' Examples of equivalent inputs include:
#'
#' \preformatted{
#' 20181012130900
#' "2018-10-12-13-09-00"
#' "2018 Oct. 12 13:09:00"
#' }
#'
#' All incoming datetimes are interpreted in the specified `timezone`. If
#' `datetime` is already `POSIXct`, it is converted to the requested timezone
#' with [lubridate::with_tz()].
#'
#' If a character datetime includes signed offset information, such as
#' `"-07:00"`, that offset is used by [lubridate::parse_date_time()] when
#' determining the equivalent instant.
#'
#' @param datetime Vector of character, numeric, integer, or `POSIXct`
#'   datetimes.
#' @param timezone Olson timezone used to interpret incoming datetimes.
#' @param expectAll Logical value specifying whether to stop if any non-missing
#'   input values fail to parse.
#' @param isJulian Logical value specifying whether `datetime` should be
#'   interpreted as a Julian date using day-of-year notation.
#' @param quiet Logical value passed to [lubridate::parse_date_time()] to
#'   suppress parsing warnings.
#'
#' @return
#' A `POSIXct` vector.
#'
#' @section Mazama Science conventions:
#' Within Mazama Science packages, datetimes not already in `POSIXct` format are
#' often represented as compact decimal values with no separators, such as
#' `20181012` or `20181012130900`, either as numbers or strings.
#'
#' @section Implementation:
#' `parseDatetime()` is a wrapper around [lubridate::parse_date_time()] that
#' defines the datetime formats supported by MazamaCoreUtils.
#'
#' @seealso
#' [lubridate::parse_date_time()]
#'
#' @examples
#' # All Y[mdHMS] formats are accepted
#' parseDatetime(2018, timezone = "America/Los_Angeles")
#' parseDatetime(201808, timezone = "America/Los_Angeles")
#' parseDatetime(20180807, timezone = "America/Los_Angeles")
#' parseDatetime(2018080718, timezone = "America/Los_Angeles")
#' parseDatetime(201808071812, timezone = "America/Los_Angeles")
#' parseDatetime(20180807181215, timezone = "America/Los_Angeles")
#'
#' parseDatetime("2018-08-07 18:12:15", timezone = "America/Los_Angeles")
#' parseDatetime("2018-08-07 18:12:15-07:00", timezone = "UTC")
#'
#' # Julian days are accepted
#' parseDatetime(
#'   2018219181215,
#'   timezone = "America/Los_Angeles",
#'   isJulian = TRUE
#' )
#'
#' # Mixed vector inputs are accepted
#' parseDatetime(
#'   c("2018-10-24 12:00", "201810311200", "2018-11-07 12:00"),
#'   timezone = "America/New_York"
#' )
#'
#' badInput <- c("20181013", NA, "20181015", "181016", "10172018")
#'
#' # Return NA for dates that cannot be parsed
#' parseDatetime(badInput, timezone = "UTC", expectAll = FALSE)
#'
#' \dontrun{
#' # Fail if any non-missing dates cannot be parsed
#' parseDatetime(badInput, timezone = "UTC", expectAll = TRUE)
#' }
#'
#' @export
#'
parseDatetime <- function(
  datetime = NULL,
  timezone = NULL,
  expectAll = FALSE,
  isJulian = FALSE,
  quiet = TRUE
) {

  # Validate parameters --------------------------------------------------------

  stopIfNull(datetime)
  stopIfNull(timezone)
  stopIfNull(expectAll)
  stopIfNull(isJulian)
  stopIfNull(quiet)

  if ( !is.character(timezone) || length(timezone) > 1 )
    stop(paste0("argument 'timezone' must be a character string of length one"))

  if ( !timezone %in% base::OlsonNames() )
    stop(sprintf("'timezone = %s' is not found in OlsonNames()", timezone))

  if ( !is.logical(expectAll) || length(expectAll) != 1 )
    stop("argument 'expectAll' must be a logical value of length one")

  if ( !is.logical(isJulian) || length(isJulian) != 1 )
    stop("argument 'isJulian' must be a logical value of length one")

  # Return early if already POSIXct -----------------------------------------

  if (lubridate::is.POSIXct(datetime))
    return(lubridate::with_tz(datetime, tzone = timezone))

  # Parse datetimes ---------------------------------------------------------

  if ( isJulian ) {

    # NOTE:  Julian date strings created by NASA satellite products often
    # NOTE:  include a digit for fractional seconds but no "." decimal marker.
    # NOTE:  We test for and fix that here.

    # Convert possible integers to character and separate fractional seconds
    datetime <- as.character(datetime)
    wholePart <- stringr::str_sub(datetime, 1, 13)
    fractionalPart <- stringr::str_sub(datetime, 14, -1)

    # Corrected date strings
    datetime <-
      paste0(wholePart, ".", fractionalPart) %>%
      stringr::str_replace("\\.$", "")

    orders <- c("Y", "Yj", "YjH", "YjHM", "YjHMS")
    parsedDatetime <- lubridate::parse_date_time(datetime,
                                                 orders,
                                                 tz = timezone,
                                                 quiet = quiet)

  } else {

    orders <- c("Y", "Ym", "Ymd", "YmdH", "YmdHM", "YmdHMS", "YmdHz", "YmdHMz", "YmdHMSz")
    parsedDatetime <- lubridate::parse_date_time(datetime,
                                                 orders,
                                                 tz = timezone,
                                                 quiet = quiet)

  }

  # Handle results ----------------------------------------------------------

  if ( all(is.na(parsedDatetime)) ) {
    stop("no datetimes could be parsed")
  }

  if ( expectAll ) {

    ## NAs that appear in the parsed datetimes and not in the original datetimes
    #  are datetimes that failed to parse (ie, not originally NA).
    failedIndices <- setdiff(
      which(is.na(parsedDatetime)),
      which(is.na(datetime))
    )

    ## If there already exist NAs in datetime, we don't want to accidently fail
    #  if all non-NA values were parsed
    if (length(failedIndices) == 1) {
      stop(paste0(
        "1 datetime failed to parse (at index: ", failedIndices, ")"
      ))

    # account for differences in plural spellings
    } else if (length(failedIndices) > 1) {
      stop(paste0(
        length(failedIndices), " datetimes failed to parse (at indices: ",
        paste0(failedIndices, collapse = ", "), ")"
      ))
    }
  }

  return(parsedDatetime)

}

#' MazamaCoreUtils
#'
#' Utility functions for production R code.
#'
#' MazamaCoreUtils provides a standardized set of utility functions used by
#' MazamaScience packages and R-based operational systems.
#'
#' Functionality includes:
#'
#' \itemize{
#'   \item Python-style logging
#'   \item simple error messaging
#'   \item cache management
#'   \item API key handling
#'   \item date parsing and formatting
#'   \item longitude/latitude validation
#'   \item unique location ID creation
#'   \item source code linting
#' }
#'
#' @keywords internal
"_PACKAGE"

# ----- Internal Package State -------------------------------------------------

MazamaCoreUtilsEnv <- new.env(parent = emptyenv())
MazamaCoreUtilsEnv$dataDir <- NULL
MazamaCoreUtilsEnv$APIKeys <- list()

# ----- API Keys ---------------------------------------------------------------

#' API keys for data services
#'
#' Internal session state used to store API keys for web services.
#'
#' Users can set API keys with [setAPIKey()]. Keys are remembered for the
#' duration of the R session and can be retrieved with [getAPIKey()].
#'
#' This provides a small abstraction layer for dependent packages so that data
#' access functions can test for and retrieve provider-specific API keys with
#' generic code.
#'
#' @name APIKeys
#' @docType data
#' @keywords environment
#'
#' @format
#' A named list of character strings.
#'
#' @seealso
#' [getAPIKey()], [setAPIKey()], [showAPIKeys()]
NULL

#' Show API keys
#'
#' Print all currently set API keys.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [getAPIKey()], [setAPIKey()]
#'
#' @keywords environment
#' @export
#' @importFrom utils str
showAPIKeys <- function() {
  utils::str(MazamaCoreUtilsEnv$APIKeys)
}

#' Get API key
#'
#' Return the API key associated with a web service provider.
#'
#' If `provider = NULL`, all currently stored API keys are returned.
#'
#' @param provider Web service provider.
#'
#' @return
#' API key string, `NULL`, or a named list of all provider/key pairs.
#'
#' @seealso
#' [APIKeys], [setAPIKey()], [showAPIKeys()]
#'
#' @keywords environment
#' @export
getAPIKey <- function(provider = NULL) {
  if ( is.null(provider) ) {
    return(MazamaCoreUtilsEnv$APIKeys)
  } else {
    return(MazamaCoreUtilsEnv$APIKeys[[provider]])
  }
}

#' Set API key
#'
#' Set the API key associated with a web service provider.
#'
#' API keys are stored in package session state and are remembered only for the
#' duration of the current R session.
#'
#' @param provider Web service provider.
#' @param key API key.
#'
#' @return
#' Invisibly returns the previous value of the API key.
#'
#' @seealso
#' [getAPIKey()], [showAPIKeys()]
#'
#' @keywords environment
#' @export
setAPIKey <- function(provider = NULL, key = NULL) {
  old <- MazamaCoreUtilsEnv$APIKeys[[provider]]
  MazamaCoreUtilsEnv$APIKeys[[provider]] <- key
  return(invisible(old))
}

test_that("set and get work", {

  # Default to NULL
  expect_identical(getAPIKey("provider1"), NULL)

  # basic set
  setAPIKey("provider1", "key1")
  setAPIKey("provider2", "key2")

  # basic get
  expect_identical(getAPIKey("provider1"), "key1")
  expect_identical(getAPIKey("provider2"), "key2")

  # show keys
  expect_identical(showAPIKeys(), str(list(provider1 = "key1", provider2 = "key2")))

  # old key is returned
  expect_identical(setAPIKey("provider1", "update1"), "key1")
  expect_identical(getAPIKey("provider1"), "update1")

})

#' Run package checks
#'
#' Convenience wrappers around [devtools::check()] for package checking at
#' different levels of thoroughness.
#'
#' These functions make it easy to run quick checks during active development
#' and more thorough checks before merging or releasing package changes.
#'
#' @param pkg Package location passed to [devtools::check()].
#'
#' @return
#' Invisibly returns the result from [devtools::check()].
#'
#' @details
#' The functions are ordered from most thorough to fastest:
#'
#' \describe{
#'   \item{`check_slowest()`}{
#'     Builds the manual, runs `donttest` and `dontrun` examples, and
#'     uses `--use-gct`.
#'   }
#'   \item{`check_slower()`}{
#'     Builds the manual and runs `donttest` and `dontrun` examples.
#'   }
#'   \item{`check_slow()`}{
#'     Builds the manual and runs `donttest` examples.
#'   }
#'   \item{`check()`}{
#'     Standard development check without building the manual or running
#'     `donttest` examples.
#'   }
#'   \item{`check_fast()`}{
#'     Skips vignette building and ignores vignettes during checking.
#'   }
#'   \item{`check_faster()`}{
#'     Skips vignette building, ignores vignettes, and skips examples.
#'   }
#'   \item{`check_fastest()`}{
#'     Skips vignette building, ignores vignettes, skips examples, and skips
#'     tests.
#'   }
#' }
#'
#' @seealso
#' [devtools::check()]
#'
#' @name packageCheck
#' @aliases check_slow check
NULL

#' @rdname packageCheck
#' @export
check <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = FALSE,
    run_dont_test = FALSE
  )

}

#' @rdname packageCheck
#' @export
check_fast <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = FALSE,
    run_dont_test = FALSE,
    build_args = c("--no-build-vignettes"),
    args = c("--ignore-vignettes")
  )

}

#' @rdname packageCheck
#' @export
check_faster <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = FALSE,
    run_dont_test = FALSE,
    build_args = c("--no-build-vignettes"),
    args = c("--ignore-vignettes", "--no-examples")
  )

}

#' @rdname packageCheck
#' @export
check_fastest <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = FALSE,
    run_dont_test = FALSE,
    build_args = c("--no-build-vignettes"),
    args = c("--ignore-vignettes", "--no-tests",
             "--no-examples")
  )

}

#' @rdname packageCheck
#' @export
check_slow <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = TRUE,
    run_dont_test = TRUE,
  )

}

#' @rdname packageCheck
#' @export
check_slower <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = TRUE,
    run_dont_test = TRUE,
    args = c("--run-dontrun")
  )

}

#' @rdname packageCheck
#' @export
check_slowest <- function(
  pkg = "."
) {

  devtools::check(
    pkg = pkg,
    manual = TRUE,
    run_dont_test = TRUE,
    args = c("--run-dontrun", "--use-gct")
  )

}

#' Create a POSIXct date range
#'
#' Create a two-element `POSIXct` vector representing a date/time range in a
#' specified timezone.
#'
#' The returned range is ordered from earliest to latest. The first element
#' represents the beginning of the requested date range and the second element
#' represents the end of the requested date range at the requested temporal
#' precision.
#'
#' By default, the returned end time is one `unit` *before* the beginning of
#' `enddate`. For example:
#'
#' \preformatted{
#' dateRange(20190101, 20190102, timezone = "UTC")
#' [1] "2019-01-01 00:00:00 UTC"
#' [2] "2019-01-01 23:59:59 UTC"
#' }
#'
#' Setting `ceilingEnd = TRUE` includes the entirety of `enddate`:
#'
#' \preformatted{
#' dateRange(
#'   20190101,
#'   20190101,
#'   timezone = "UTC",
#'   ceilingEnd = TRUE
#' )
#' [1] "2019-01-01 00:00:00 UTC"
#' [2] "2019-01-01 23:59:59 UTC"
#' }
#'
#' The `ceilingEnd` argument addresses ambiguity in phrases such as
#' `"August 1-8"`. With `ceilingEnd = FALSE` (default), the range extends
#' through the end of August 7, stopping at the midnight boundary where August 8
#' begins. With `ceilingEnd = TRUE`, the range
#' extends through the end of August 8.
#'
#' Input dates are parsed with [parseDatetime()] using the specified
#' `timezone`.
#'
#' @param startdate Desired start datetime.
#' @param enddate Desired end datetime.
#' @param timezone Olson timezone used to interpret incoming dates.
#' @param unit Temporal precision used for the returned end-of-range value.
#'   One of `"day"`, `"hour"`, `"min"`, or `"sec"`.
#' @param ceilingStart Logical specifying whether to round `startdate` up to the
#'   next day boundary instead of down.
#' @param ceilingEnd Logical specifying whether to include the entirety of the
#'   final day.
#' @param days Number of days to include when either `startdate` or `enddate`
#'   is omitted.
#'
#' @return
#' Two-element `POSIXct` vector ordered from earliest to latest.
#'
#' @section Default arguments:
#' If either `startdate` or `enddate` is missing, the missing boundary is
#' calculated using `days`.
#'
#' If both are missing, `enddate` defaults to the current day in `timezone`
#' and `startdate` is calculated as `enddate - days`.
#'
#' @section End-of-day units:
#' The returned end time is adjusted to the last representable value within the
#' requested unit:
#'
#' \describe{
#'   \item{`unit = "day"`}{End time is midnight at the start of the final day.}
#'   \item{`unit = "hour"`}{End time is `23:00:00`.}
#'   \item{`unit = "min"`}{End time is `23:59:00`.}
#'   \item{`unit = "sec"`}{End time is `23:59:59`.}
#' }
#'
#' @section POSIXct inputs:
#' When `startdate` or `enddate` are already `POSIXct` values, they are first
#' converted to `timezone` with [lubridate::with_tz()] without changing the
#' represented instant in time.
#'
#' @section Parameter precedence:
#' When parameters conflict, the following rules apply:
#'
#' \enumerate{
#'   \item If both `startdate` and `enddate` are supplied, `days` is ignored.
#'   \item If `startdate` is missing, `ceilingStart` is ignored.
#'   \item If `enddate` is missing, `ceilingEnd` is ignored.
#' }
#'
#' @examples
#' dateRange("2019-01-08", timezone = "UTC")
#'
#' dateRange("2019-01-08", unit = "min", timezone = "UTC")
#'
#' dateRange("2019-01-08", unit = "hour", timezone = "UTC")
#'
#' dateRange("2019-01-08", unit = "day", timezone = "UTC")
#'
#' dateRange("2019-01-08", "2019-01-11", timezone = "UTC")
#'
#' dateRange(
#'   enddate = 20190112,
#'   days = 3,
#'   unit = "day",
#'   timezone = "America/Los_Angeles"
#' )
#'
#' @export
#'
dateRange <- function(
  startdate = NULL,
  enddate = NULL,
  timezone = NULL,
  unit = "sec",
  ceilingStart = FALSE,
  ceilingEnd = FALSE,
  days = 7
) {

  # Validate parameters --------------------------------------------------------

  stopIfNull(timezone)
  stopIfNull(unit)
  ceilingStart <- setIfNull(ceilingStart, FALSE, "logical")
  ceilingEnd <- setIfNull(ceilingEnd, FALSE, "logical")

  if ( !timezone %in% base::OlsonNames() )
    stop(sprintf("'timezone = %s' is not found in OlsonNames()", timezone))

  if ( !is.numeric(days) || length(days) > 1 || days < 1 )
    stop("'days' must be a single positive number")

  if ( !is.null(startdate) && length(startdate) != 1 )
    stop("'startdat'e must be of length one, if specified")

  if ( !is.null(enddate) && length(enddate) != 1 )
    stop("'enddate' must be of length one, if specified")


  # Handle end-of-day unit -----------------------------------------------------

  if ( stringr::str_detect(unit, "^day") ) {
    endUnitAdjust <- lubridate::days(0)
  } else if ( stringr::str_detect(unit, "^hour") ) {
    endUnitAdjust <- lubridate::hours(1)
  } else if ( stringr::str_detect(unit, "^min") ) {
    endUnitAdjust <- lubridate::minutes(1)
  } else if ( stringr::str_detect(unit, "^sec") ) {
    endUnitAdjust <- lubridate::seconds(1)
  } else {
    stop("'unit' must be one of: 'day', 'hour', 'min', 'sec'")
  }

  # Determine start and end times ----------------------------------------------

  # NOTE:  Always assume floor_date for both start- and enddate and take
  # NOTE:  care of ceilingEnd as the very last step

  if ( !is.null(startdate) && !is.null(enddate) ) {

    # ** Both found: use startdate, enddate ------------------------------------

    # Handle parsing and ordering
    timeInput <- timeRange(startdate, enddate, timezone = timezone)

    if ( ceilingStart ) {
      starttime <-
        timeInput[1] %>%
        lubridate::ceiling_date(unit = "day")
    } else {
      starttime <-
        timeInput[1] %>%
        lubridate::floor_date(unit = "day")
    }

    endtime <-
      timeInput[2] %>%
      lubridate::floor_date(unit = "day")

    if ( starttime == endtime ) {
      # Handle single-day ranges
      endtime <- endtime + lubridate::ddays(1)
    } else {
      # Adjust for ceilingEnd
      if ( ceilingEnd ) {
        endtime <- endtime + lubridate::ddays(1)
      }
    }

    endtime <- endtime - endUnitAdjust

  } else if ( !is.null(startdate) && is.null(enddate) ) {

  # ** Missing enddate: use startdate, (startdate + days) ----------------------

    if ( ceilingStart ) {
      starttime <-
        startdate %>%
        parseDatetime(timezone = timezone) %>%
        lubridate::ceiling_date(unit = "day")
    } else {
      starttime <-
        startdate %>%
        parseDatetime(timezone = timezone) %>%
        lubridate::floor_date(unit = "day")
    }

    endtime <- starttime + lubridate::days(days)

    # Ignore "ceilingEnd" parameter

    endtime <- endtime - endUnitAdjust

  } else if ( is.null(startdate) && !is.null(enddate) ) {

    # ** Missing startdate: use (enddate - days), enddate ----------------------

    endtime <-
      enddate %>%
      parseDatetime(timezone = timezone) %>%
      lubridate::floor_date(unit = "day")

    if ( ceilingEnd ) {
      endtime <- endtime + lubridate::ddays(1)
    }

    # Ignore ceilingStart

    starttime <- endtime - lubridate::days(days)

    endtime <- endtime - endUnitAdjust


  } else {

    # ** Both missing: use (now - days), now -----------------------------------

    endtime <-
      lubridate::now(tzone = timezone) %>%
      lubridate::floor_date(unit = "day")

    if ( ceilingEnd ) {
      endtime <- endtime + lubridate::ddays(1)
    }

    # Ignore ceilingStart

    starttime <- endtime - lubridate::days(days)

    endtime <- endtime - endUnitAdjust

  }


  # Return tlim ----------------------------------------------------------------

  return(c(starttime, endtime))

}

test_that("dateRange() requires explicit timezone", {

  expect_error(dateRange(timezone = NULL))

})

test_that("Output is a two element POSIXct vector", {


  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  tlim <- dateRange(startdate = startTime, enddate = endTime, timezone = "America/Los_Angeles")

  expect_length(tlim, 2)
  expect_s3_class(tlim, "POSIXct")

})


test_that("Output has the correct timezone", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  tlim_LA <- dateRange(startdate = startTime, enddate = endTime, timezone = "America/Los_Angeles")
  tlim_UTC <- dateRange(startdate = startTime, enddate = endTime, timezone = "UTC")

  expect_identical(attributes(tlim_LA)$tzone, "America/Los_Angeles")
  expect_identical(attributes(tlim_UTC)$tzone, "UTC")

})


test_that("Equivalent inputs create the same output", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")
  dayCount <- 7

  tlim_startEnd <- dateRange(startdate = startTime, enddate = endTime, timezone = "America/Los_Angeles")
  tlim_startDays <- dateRange(startdate = startTime, days = dayCount, timezone = "America/Los_Angeles")
  tlim_endDays <- dateRange(enddate = endTime, days = dayCount, timezone = "America/Los_Angeles")

  expect_identical(tlim_startEnd, tlim_startDays)
  expect_identical(tlim_startEnd, tlim_endDays)
  expect_equal(tlim_startDays, tlim_endDays)

})


test_that("First element in output is less than the second", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  tlim <- dateRange(startdate = startTime, enddate = endTime, timezone = "America/Los_Angeles")
  tlim_now <- dateRange(timezone = "America/Los_Angeles")
  tlim_rev <- dateRange(startdate = endTime, enddate = startTime, timezone = "America/Los_Angeles")

  expect_lt(tlim[1], tlim[2])
  expect_lt(tlim_now[1], tlim_now[2])
  expect_lt(tlim_rev[1], tlim_rev[2])

})


test_that("POSIXct inputs retain their point-in-time", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  startTimeString <- strftime(startTime, format = "%Y%m%d%H%M%S", tz = "America/Los_Angeles")
  endTimeString <- strftime(endTime, format = "%Y%m%d%H%M%S", tz = "America/Los_Angeles")

  expectedStartTime_UTC <-
    startTime %>%
    lubridate::with_tz(tzone = "UTC") %>%
    lubridate::floor_date(unit = "day")

  expectedEndTime_UTC <-
    endTime %>%
    lubridate::with_tz(tzone = "UTC") %>%
    lubridate::floor_date(unit = "day") %>%
    `-`(lubridate::dseconds(1))


  # Test with POSIXct inputs for start/end
  tlim <- dateRange(startdate = startTime, enddate = endTime, timezone = "UTC", unit = "sec")

  expect_identical(tlim[1], expectedStartTime_UTC)
  expect_identical(tlim[2], expectedEndTime_UTC)


  # Test with POSIXct input for start
  tlim <- dateRange(startdate = startTime, enddate = endTimeString, timezone = "UTC", unit = "sec")

  expect_identical(tlim[1], expectedStartTime_UTC)


  # Test with POSIXct input for end
  tlim <- dateRange(startdate = startTimeString, enddate = endTime, timezone = "UTC", unit = "sec")

  expect_identical(tlim[2], expectedEndTime_UTC)

})


test_that("Non-POSIXct inputs work as expected", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  startTime_string <- strftime(startTime, format = "%Y-%m-%d %H:%M:%S", tz = "America/Los_Angeles")
  starTime_numeric <- strftime(startTime, format = "%Y%m%d%H%M%S", tz = "America/Los_Angeles") %>% as.numeric()

  endTime_string <- strftime(endTime, format = "%Y-%m-%d %H:%M:%S", tz = "America/Los_Angeles")
  endTime_numeric <- strftime(endTime, format = "%Y%m%d%H%M%S", tz = "America/Los_Angeles") %>% as.numeric()

  tlim_posixct <- dateRange(startdate = startTime, enddate = endTime, timezone = "America/Los_Angeles")
  tlim_string <- dateRange(startdate = startTime_string, enddate = endTime_string, timezone = "America/Los_Angeles")
  tlim_numeric <- dateRange(startdate = starTime_numeric, enddate = endTime_numeric, timezone = "America/Los_Angeles")

  expect_identical(tlim_posixct, tlim_string)
  expect_identical(tlim_posixct, tlim_numeric)

})


test_that("End-of-Day unit works as expected", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  # Standard input
  tlim_sec <- dateRange(startdate = startTime, enddate = endTime, timezone = "UTC", unit = "sec")
  tlim_min <- dateRange(startdate = startTime, enddate = endTime, timezone = "UTC", unit = "min")
  tlim_hour <- dateRange(startdate = startTime, enddate = endTime, timezone = "UTC", unit = "hour")
  tlim_day <- dateRange(startdate = startTime, enddate = endTime, timezone = "UTC", unit = "day")

  expect_identical(strftime(tlim_sec[2], format = "%H%M%S", tz = "UTC"), "235959")
  expect_identical(strftime(tlim_min[2], format = "%H%M%S", tz = "UTC"), "235900")
  expect_identical(strftime(tlim_hour[2], format = "%H%M%S", tz = "UTC"), "230000")
  expect_identical(strftime(tlim_day[2], format = "%H%M%S", tz = "UTC"), "000000")

  # Reversed input
  rev_tlim <- dateRange(startdate = endTime, enddate = startTime, timezone = "UTC", unit = "sec")

  expect_identical(strftime(rev_tlim[2], format = "%H%M%S", tz = "UTC"), "235959")

})


test_that("Default arguments work as expected", {

  startTime <- ISOdatetime(2019, 08, 1, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  numDays <- 7

  expectedStart <- lubridate::floor_date(startTime, unit = "day")
  expectedEnd <- lubridate::floor_date(endTime, unit = "day")

  expect_identical(
    dateRange(
      startdate = startTime, enddate = NULL,
      timezone = "America/Los_Angeles", unit = "day"
    ),
    c(expectedStart, expectedEnd)
  )

  expect_identical(
    dateRange(
      startdate = NULL, enddate = endTime,
      timezone = "America/Los_Angeles", unit = "day"
    ),
    c(expectedStart, expectedEnd)
  )

  expect_identical(
    dateRange(
      startdate = NULL, enddate = NULL,
      timezone = "America/Los_Angeles", unit = "day"
    ),
    c(
      lubridate::now(tzone = "America/Los_Angeles") %>%
        lubridate::floor_date(unit = "day") %>%
        `-`(lubridate::days(numDays)),
      lubridate::now(tzone = "America/Los_Angeles") %>%
        lubridate::floor_date(unit = "day")
    )
  )

})

test_that("unit, ceilingStart and ceilingEnd arguments work", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  startDate <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")
  endDate <- ISOdatetime(2019, 08, 7, 23, 59, 59, tz = "America/Los_Angeles")

  # NOTE:  start always begins at midnight regarless of unit
  startMin <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")
  startHour <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")
  startDay <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")

  startMinCeil <- ISOdatetime(2019, 08, 02, 00, 00, 00, tz = "America/Los_Angeles")
  startHourCeil <- ISOdatetime(2019, 08, 02, 00, 00, 00, tz = "America/Los_Angeles")
  startDayCeil <- ISOdatetime(2019, 08, 02, 00, 00, 00, tz = "America/Los_Angeles")

  endMin <- ISOdatetime(2019, 08, 7, 23, 59, 00, tz = "America/Los_Angeles")
  endHour <- ISOdatetime(2019, 08, 7, 23, 00, 00, tz = "America/Los_Angeles")
  endDay <- ISOdatetime(2019, 08, 8, 00, 00, 00, tz = "America/Los_Angeles")

  endMinCeil <- ISOdatetime(2019, 08, 08, 23, 59, 00, tz = "America/Los_Angeles")
  endHourCeil <- ISOdatetime(2019, 08, 08, 23, 00, 00, tz = "America/Los_Angeles")
  endDayCeil <- ISOdatetime(2019, 08, 09, 00, 00, 00, tz = "America/Los_Angeles")

  tlim <- dateRange(startTime, endTime, timezone = "America/Los_Angeles")

  tlimMin <- dateRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "min")
  tlimHour <- dateRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "hour")
  tlimDay <- dateRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "day")

  tlimMinCeil <- dateRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "min", ceilingStart = TRUE, ceilingEnd = TRUE)
  tlimHourCeil <- dateRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "hour", ceilingStart = TRUE, ceilingEnd = TRUE)
  tlimDayCeil <- dateRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "day", ceilingStart = TRUE, ceilingEnd = TRUE)

  expect_identical(tlim, c(startDate, endDate))
  expect_identical(tlimMin, c(startMin, endMin))
  expect_identical(tlimHour, c(startHour, endHour))
  expect_identical(tlimDay, c(startDay, endDay))
  expect_identical(tlimMinCeil, c(startMinCeil, endMinCeil))
  expect_identical(tlimHourCeil, c(startHourCeil, endHourCeil))
  expect_identical(tlimDayCeil, c(startDayCeil, endDayCeil))

})

test_that("days are honored when start or endtime is missing", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 08, 20, 30, 40, tz = "America/Los_Angeles")

  startDate <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")
  endDate <- ISOdatetime(2019, 08, 7, 23, 59, 59, tz = "America/Los_Angeles")
  endDateCeil <- ISOdatetime(2019, 08, 8, 23, 59, 59, tz = "America/Los_Angeles")

  # Normal usage
  tlim <- dateRange(startTime, endTime,
                    timezone = "America/Los_Angeles")
  expect_identical(7, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

  # Full end enddate
  tlim <- dateRange(startTime, endTime, ceilingEnd = TRUE,
                    timezone = "America/Los_Angeles")
  expect_identical(8, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

  # Ignore days when both start and end are specified
  tlim <- dateRange(startTime, endTime, days = 5,
                    timezone = "America/Los_Angeles")
  expect_identical(7, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

  tlim <- dateRange(startTime, endTime, ceilingEnd = TRUE,
                    timezone = "America/Los_Angeles")
  expect_identical(8, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

  # Honor days when startdate is missing
  tlim <- dateRange(enddate = endTime, days = 3,
                    timezone = "America/Los_Angeles")
  expect_identical(3, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

  # Honor both days and ceilingEnd, adjusting starttime
  tlim <- dateRange(enddate = endTime, ceilingEnd = TRUE, days = 3,
                    timezone = "America/Los_Angeles")
  expect_identical(3, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))
  expect_identical(endDateCeil, tlim[2])

  # Honor days when enddate is missing
  tlim <- dateRange(startdate = startTime, days = 3,
                    timezone = "America/Los_Angeles")
  expect_identical(3, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

  # Honor days but ignore ceilingEnd when enddate is not specified
  tlim <- dateRange(startdate = startTime, ceilingEnd = TRUE, days = 3,
                    timezone = "America/Los_Angeles")
  expect_identical(3, round(as.numeric(difftime(tlim[2], tlim[1], units="days"))))

})

test_that("Single day ranges are handled propery", {

  startTime <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 01, 23, 59, 59, tz = "America/Los_Angeles")

  tlim <- dateRange(startdate = startTime, enddate = endTime,
                    timezone = "America/Los_Angeles")
  expect_identical( c(startTime, endTime), tlim )

  # First example from function description
  tlim <- dateRange(20190801, 20190802,
                    timezone = "America/Los_Angeles")
  expect_identical( c(startTime, endTime), tlim )

  # Second example from function description
  tlim <- dateRange(20190801, 20190801, ceilingEnd = TRUE,
                    timezone = "America/Los_Angeles")
  expect_identical( c(startTime, endTime), tlim )

})

#' Extract tables from an HTML page
#'
#' Parse an HTML page and return all `<table>` elements as a list of data
#' frames.
#'
#' The `url` argument may be either a remote URL or a local file path. Tables are
#' parsed with [rvest::html_table()]. To extract a single table, use
#' [html_getTable()].
#'
#' @param url URL or local file path of an HTML page.
#' @param header Logical specifying whether the first row should be used as
#'   column names. If `NA`, the first row is used only when it contains `<th>`
#'   elements.
#'
#' @return
#' List of data frames, one for each HTML table.
#'
#' @examples
#' \dontrun{
#' url <- "https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
#'
#' tables <- html_getTables(url)
#' firstTable <- tables[[1]]
#'
#' head(firstTable)
#' nrow(firstTable)
#' }
#'
#' @name html_getTables
#' @rdname html_getTables
#' @export

html_getTables <- function(
  url = NULL,
  header = NA
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(url)

  # ----- Extract the tables ----------------------------------------------------

  try({

    # Get the raw HTML from the URL
    urlXML <- xml2::read_html(url)

    # Get a list of tables in the document
    tableNodes <- rvest::html_elements(urlXML, css = "table")

    # Make this list human-readable
    tables_clean <- rvest::html_table(
      tableNodes,
      header = header,        # use first row as header
      trim = TRUE,            # remove leading/trailing white space
      fill = TRUE,            # NA fill rows with fewer than max columns
      dec = "."
    )

  }, silent = TRUE) %>%
  stopOnError()

  # ----- Return ---------------------------------------------------------------

  return(tables_clean)

}

#' @rdname html_getTables
#'
#' @param index Index identifying which table to return.
#'
#' @return
#' A single data frame containing the requested HTML table.
#'
#' @export
html_getTable <- function(
  url = NULL,
  header = NA,
  index = 1
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(url)

  index <- as.integer(index)
  if ( index < 1 )
    index <- 1

  # ----- Extract the table ----------------------------------------------------

  try({

    # Get a list of tables in this document
    tables <- html_getTables(url, header = header)

    returnTable <- tables[[index]]

  }, silent = TRUE) %>%
  stopOnError()

  # ----- Return ---------------------------------------------------------------

  return(returnTable)

}

#' Create a POSIXct date sequence
#'
#' Create a sequence of local-midnight `POSIXct` datetimes in a specified
#' timezone.
#'
#' The returned sequence begins at midnight local time on `startdate` and ends
#' at midnight local time on `enddate`, *i.e.* the *beginning* of `enddate`.
#'
#' The `ceilingEnd` argument addresses ambiguity in date ranges such as
#' `"August 1-8"`. With `ceilingEnd = FALSE` (default), the sequence ends at
#' the beginning of August 8. With `ceilingEnd = TRUE`, the sequence
#' includes the entirety of August 8, ending at the midnight that begins August 9.
#'
#' Input dates are parsed with [parseDatetime()] using the specified
#' `timezone`. Any hour-minute-second information is removed after parsing.
#'
#' @note
#' This function preserves local clock-time midnight boundaries across daylight
#' savings transitions. This differs from `seq.Date(..., by = "day")`, which
#' advances by fixed 24-hour intervals and can drift away from midnight local
#' time during daylight savings changes.
#'
#' @param startdate Desired start datetime.
#' @param enddate Desired end datetime.
#' @param timezone Olson timezone used to interpret incoming dates.
#' @param ceilingEnd Logical specifying whether to include the end of the final
#'   day.
#'
#' @return
#' A vector of `POSIXct` datetimes at local midnight.
#'
#' @section POSIXct inputs:
#' When `startdate` or `enddate` are already `POSIXct` values, they are first
#' converted to `timezone` with [lubridate::with_tz()] without changing the
#' represented instant in time. They are then floored to local midnight.
#'
#' @examples
#' dateSequence(
#'   "2019-11-01",
#'   "2019-11-08",
#'   timezone = "America/Los_Angeles"
#' )
#'
#' dateSequence(
#'   "2019-11-01",
#'   "2019-11-07",
#'   timezone = "America/Los_Angeles",
#'   ceilingEnd = TRUE
#' )
#'
#' # Observe daylight savings handling
#' datetime <- dateSequence(
#'   "2019-11-01",
#'   "2019-11-08",
#'   timezone = "America/Los_Angeles"
#' )
#'
#' datetime
#' lubridate::with_tz(datetime, "UTC")
#'
#' # POSIXct inputs preserve the represented instant before flooring
#' jst <- dateSequence(
#'   20190307,
#'   20190315,
#'   timezone = "Asia/Tokyo"
#' )
#'
#' jst
#'
#' dateSequence(
#'   jst[1],
#'   jst[7],
#'   timezone = "UTC"
#' )
#'
#' @export
#'
dateSequence <- function(
  startdate = NULL,
  enddate = NULL,
  timezone = NULL,
  ceilingEnd = FALSE
) {

  # ----- Validate parameters --------------------------------------------------

  stopIfNull(startdate)
  stopIfNull(enddate)
  stopIfNull(timezone)
  stopIfNull(ceilingEnd)

  if ( !timezone %in% base::OlsonNames() )
    stop(sprintf("'timezone = %s' is not found in OlsonNames()", timezone))

  # ----- Create sequence ------------------------------------------------------

  # NOTE:  Need to interpret date in local time first, then floor, then go to
  # NOTE:  the middle of the day. Otherwise strftime() below will repeat the
  # NOTE:  date on which we switch from daylight savings to standard time.
  start <-
    parseDatetime(startdate, timezone = timezone) %>%
    lubridate::floor_date(unit = "day") + lubridate::dhours(12)

  end <-
    parseDatetime(enddate, timezone = timezone) %>%
    lubridate::floor_date(unit = "day") + lubridate::dhours(12)

  if ( ceilingEnd )
    end <- end + lubridate::ddays(1)

  # NOTE:  seq.Date(..., by = "day") operates by repeatedly adding 24 hours
  # NOTE:  which means that when we switch to/from daylight savings we end up
  # NOTE:  no longer on the midnight local time day boundary. Hence the
  # NOTE:  following workaround

  datetime <-
    seq(start, end, by = "day") %>%
    strftime("%Y%m%d", tz = timezone) %>%
    MazamaCoreUtils::parseDatetime(timezone = timezone)

  # ----- Return ---------------------------------------------------------------

  return(datetime)

}

test_that("explicit arguments are required", {

  expect_error(dateSequence(startdate = NULL))
  expect_error(dateSequence(enddate = NULL))
  expect_error(dateSequence(timezone = NULL))
  expect_error(dateSequence(ceilingEnd = NULL))

})

test_that("daylight savings is handled and ceilingEnd works", {

  startdate <- 20191101
  enddate <- 20191107

  localDatetimes_8 <- c(
    ISOdatetime(2019, 11, 01, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 02, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 03, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 04, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 05, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 06, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 07, 00, 00, 00, tz = "America/Los_Angeles"),
    ISOdatetime(2019, 11, 08, 00, 00, 00, tz = "America/Los_Angeles")
  )

  localDatetimes_7 <- localDatetimes_8[1:7]

  datetime_LA_7 <- dateSequence(startdate = startdate, enddate = enddate,
                              timezone = "America/Los_Angeles")
  datetime_LA_8 <- dateSequence(startdate = startdate, enddate = enddate,
                              timezone = "America/Los_Angeles",
                              ceilingEnd = TRUE)

  expect_identical(datetime_LA_7, localDatetimes_7)
  expect_identical(datetime_LA_8, localDatetimes_8)

})

test_that("POSIXct preserves instant in time", {

  startdate <- 20190308
  enddate <- 20190312

  datetime_JST_5 <- c(
    ISOdatetime(2019, 03, 08, 00, 00, 00, tz = "Asia/Tokyo"),
    ISOdatetime(2019, 03, 09, 00, 00, 00, tz = "Asia/Tokyo"),
    ISOdatetime(2019, 03, 10, 00, 00, 00, tz = "Asia/Tokyo"),
    ISOdatetime(2019, 03, 11, 00, 00, 00, tz = "Asia/Tokyo"),
    ISOdatetime(2019, 03, 12, 00, 00, 00, tz = "Asia/Tokyo")
  )

  datetime_UTC_5 <- c(
    ISOdatetime(2019, 03, 07, 00, 00, 00, tz = "UTC"),
    ISOdatetime(2019, 03, 08, 00, 00, 00, tz = "UTC"),
    ISOdatetime(2019, 03, 09, 00, 00, 00, tz = "UTC"),
    ISOdatetime(2019, 03, 10, 00, 00, 00, tz = "UTC"),
    ISOdatetime(2019, 03, 11, 00, 00, 00, tz = "UTC")
  )

  # Passing in POSIXct values preserves the instant in time before flooring --
  #   midnight Tokyo time is the day before in UTC
  JST <- dateSequence(startdate, enddate, timezone = "Asia/Tokyo")
  UTC <- dateSequence(JST[1], JST[5], timezone = "UTC")

  expect_identical(JST, datetime_JST_5)
  expect_identical(UTC, datetime_UTC_5)

})

#' Create one or more unique location IDs
#'
#' Create a location ID for each longitude/latitude pair using a geohash.
#'
#' Each location ID is unique within a geohash grid cell. The `precision`
#' argument determines the size of the grid cell. At the equator, approximate
#' grid cell widths are:
#'
#' \preformatted{
#' precision   maximum grid cell width
#'         5   ~ 4.9 km
#'         6   ~ 1.2 km
#'         7   ~ 153 m
#'         8   ~ 38 m
#'         9   ~ 4.8 m
#'        10   ~ 1.2 m
#' }
#'
#' Invalid locations are assigned the value specified by `invalidID`, typically
#' `NA`.
#'
#' @param longitude Vector of longitudes in decimal degrees east.
#' @param latitude Vector of latitudes in decimal degrees north.
#' @param precision Precision used when encoding geohashes.
#' @param algorithm Encoding algorithm to use. Only `"geohash"` is currently
#'   supported. `"digest"` is deprecated and will generate an error.
#' @param invalidID Identifier to use for invalid locations. This can be a
#'   character string or `NA`.
#'
#' @return
#' Character vector of location IDs.
#'
#' @examples
#' longitude <- c(-122.5, 0, NA, -122.5, -122.5)
#' latitude <- c(47.5, 0, 47.5, NA, 47.5)
#'
#' createLocationID(longitude, latitude)
#' createLocationID(longitude, latitude, precision = 7)
#' createLocationID(longitude, latitude, invalidID = "bad")
#'
#' @references
#' <https://michaelchirico.github.io/geohashTools/index.html>
#'
#' @export
createLocationID <- function(
    longitude = NULL,
    latitude = NULL,
    precision = 10,
    algorithm = c("geohash", "digest"),
    invalidID = as.character(NA)
) {

  # Validate parameters --------------------------------------------------------

  stopIfNull(longitude)
  stopIfNull(latitude)

  precision <- setIfNull(precision, 10)

  algorithm <- match.arg(algorithm)

  if ( algorithm == "digest" ) {
    stop(
      "algorithm = 'digest' is no longer supported.\n\n",
      "Please use algorithm = 'geohash'."
    )
  }

  suppressWarnings({
    longitude <- as.numeric(longitude)
    latitude <- as.numeric(latitude)
  })


  # Create locationID ----------------------------------------------------------

  locationID <- rep(as.character(invalidID), times = length(longitude))
  mask <- createLocationMask(longitude, latitude, removeZeroZero = FALSE)

  # If all locations are bad, return immediately.
  if ( sum(mask) == 0 )
    return(locationID)

  locationID[mask] <- mapply(
    function(lat, lon) {
      geohashTools::gh_encode(lat, lon, precision)
    },
    latitude[mask],
    longitude[mask],
    SIMPLIFY = TRUE,
    USE.NAMES = FALSE
  )


  # Return ---------------------------------------------------------------------

  return(locationID)

}

test_that("createLocationID() validates input", {

  expect_error(createLocationID("abc"))
  expect_error(createLocationID(-120.325278))
  expect_error(createLocationID(1:4, 1:5))

})

test_that("createLocationID() handles invalid locations", {

  longitude <- c(-122.5, 0, NA, -122.5, -122.5)
  latitude <- c( 47.5, 0, 47.5,   NA, 47.5)
  algorithm <- "geohash"
  invalidID = "bad"

  expect_equal(
    createLocationID(longitude, latitude),
    c("c22yhrn5x1", "s000000000", NA, NA, "c22yhrn5x1")
  )

  expect_equal(
    createLocationID(longitude, latitude, invalidID = "bad"),
    c("c22yhrn5x1", "s000000000", "bad", "bad", "c22yhrn5x1")
  )

})

test_that("createLocationID() rounds properly", {

  # Wenatchee

  # Test significant digits
  expect_equal(
    createLocationID(-120.325278, 47.423333),
    "c26mvcjucy"
  )

  expect_equal(
    createLocationID(-120.3252780, 47.4233330),
    "c26mvcjucy"
  )

  # Test rounding
  expect_equal(
    createLocationID(-120.32527804, 47.42333304),
    "c26mvcjucy"
  )

  expect_equal(
    createLocationID(-120.32527796, 47.42333296),
    "c26mvcjucy"
  )

  # Different enough
  expect_false(
    createLocationID(-120.3252, 47.4233) == "c26mvcjucy"
  )

})

test_that("createLocationID() vectors work", {

  # Setup
  longitude <- -120:-110
  latitude <- 30:40

  # Default to "geohash"
  expect_identical(
    createLocationID(longitude, latitude),
    c("9m6dtm6dtm","9me2k56u54","9msn4c7j88","9mug9x7nym","9qj92me2k5","9qnpp5e9cb",
      "9qquvceepq","9qxdkxeut5","9wb25msn4c","9wcjf5sr2w","9x1g8cu2yh")
  )

})

test_that("precision is used properly", {

  # Wenatchee

  expect_equal(
    createLocationID(-120.325278, 47.423333, precision = 10),
    "c26mvcjucy"
  )

  expect_equal(
    createLocationID(-120.325278, 47.423333, precision = 9),
    "c26mvcjuc"
  )

  expect_equal(
    createLocationID(-120.325278, 47.423333, precision = 8),
    "c26mvcju"
  )

  expect_equal(
    createLocationID(-120.325278, 47.423333, precision = 7),
    "c26mvcj"
  )

  expect_equal(
    createLocationID(-120.325278, 47.423333, precision = 6),
    "c26mvc"
  )

})

#' Create a mask of valid locations
#'
#' Create a logical mask identifying valid longitude/latitude pairs.
#'
#' The returned logical vector contains `TRUE` for valid locations and `FALSE`
#' for invalid locations. This is useful for filtering data frames to retain
#' only records with valid geographic coordinates.
#'
#' Longitude and latitude values are considered valid when they:
#'
#' \itemize{
#'   \item fall within `lonRange` and `latRange`
#'   \item are not missing
#'   \item are not located at `(0, 0)` when `removeZeroZero = TRUE`
#' }
#'
#' The `lonRange` and `latRange` arguments can be used to restrict valid
#' locations to a rectangular geographic region.
#'
#' @param longitude Vector of longitudes in decimal degrees east.
#' @param latitude Vector of latitudes in decimal degrees north.
#' @param lonRange Range of valid longitudes.
#' @param latRange Range of valid latitudes.
#' @param removeZeroZero Logical specifying whether the coordinate pair
#'   `(0, 0)` should be treated as invalid.
#'
#' @return
#' Logical vector identifying valid locations.
#'
#' @examples
#' createLocationMask(
#'   longitude = c(-120, NA, -120, -220, -120, 0),
#'   latitude = c(45, 45, NA, 45, 100, 0)
#' )
#'
#' createLocationMask(
#'   longitude = -120:-90,
#'   latitude = 20:50,
#'   lonRange = c(-110, -100),
#'   latRange = c(30, 40)
#' )
#'
#' @rdname createLocationMask
#' @export
#'
createLocationMask <- function(
    longitude = NULL,
    latitude = NULL,
    lonRange = c(-180, 180),
    latRange = c(-90, 90),
    removeZeroZero = TRUE
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(longitude)
  MazamaCoreUtils::stopIfNull(latitude)
  MazamaCoreUtils::stopIfNull(lonRange)
  MazamaCoreUtils::stopIfNull(latRange)
  MazamaCoreUtils::stopIfNull(removeZeroZero)

  suppressWarnings({
    longitude <- as.numeric(longitude)
    latitude <- as.numeric(latitude)
  })

  if ( length(longitude) != length(latitude) )
    stop("longitude and latitude are not the same size.")

  if ( any(lonRange < -180) || any(lonRange > 180) )
    stop("lonRange includes invalid longitudes")

  if ( any(latRange < -90) || any(latRange > 90) )
    stop("latRange includes invalid latitudes")

  # ----- Build mask -----------------------------------------------------------

  zero_mask <- rep(TRUE, times = length(longitude))
  if ( removeZeroZero )
    zero_mask <- (longitude + latitude) != 0.0

  mask <-
    zero_mask &
    longitude >= lonRange[1] &
    longitude <= lonRange[2] &
    latitude >= latRange[1] &
    latitude <= latRange[2]

  mask[is.na(mask)] <- FALSE

  # ----- Return ---------------------------------------------------------------

  return(mask)

}

# utils-logging.R
#
# Refactor note (2025-12-31):
#   This file now uses the {logger} package under the hood while retaining the
#   MazamaCoreUtils public API:
#     - logger.setup()
#     - logger.isInitialized()
#     - logger.setLevel()
#     - logger.trace()/debug()/info()/warn()/error()/fatal()
#
#   The exported log level constants (FATAL/ERROR/WARN/INFO/DEBUG/TRACE) are
#   retained for backwards compatibility.

# ------------------------------------------------------------------------------
# Internal constants + helpers
# ------------------------------------------------------------------------------

.MAZAMA_LOG_NAMESPACE <- "MazamaCoreUtils"

# This function is conceptually "appender.null()" from futile.logger.
# Used when a given log file is NULL (disabled).
appender.null <- function() {
  function(lines) invisible(NULL)
}

# Map MazamaCoreUtils levels (and/or logger levels) to logger package constants.
.logger_map_level <- function(level) {
  # Allow passing the MazamaCoreUtils exported constants, which have names().
  if (is.numeric(level) && length(level) == 1L && !is.null(names(level))) {
    lvl_name <- names(level)[1]
  } else if (is.character(level) && length(level) == 1L) {
    lvl_name <- level
  } else if (is.numeric(level) && length(level) == 1L) {
    # Best effort: if user passes a raw integer, assume it's already a logger level.
    return(level)
  } else {
    stop("Invalid log level. Use one of: TRACE, DEBUG, INFO, WARN, ERROR, FATAL.", call. = FALSE)
  }

  lvl_name <- toupper(lvl_name)

  if (!requireNamespace("logger", quietly = TRUE)) {
    stop("Package 'logger' must be installed to use MazamaCoreUtils logging.", call. = FALSE)
  }

  # logger has OFF, FATAL, ERROR, WARN, SUCCESS, INFO, DEBUG, TRACE
  if (!exists(lvl_name, envir = asNamespace("logger"), inherits = FALSE)) {
    stop(sprintf("Unknown log level '%s'.", lvl_name), call. = FALSE)
  }

  get(lvl_name, envir = asNamespace("logger"), inherits = FALSE)
}

# Remove all existing logger indices in our namespace to make setup idempotent.
#
# IMPORTANT:
#   The {logger} package expects at least one config/index in a namespace.
#   Deleting down to zero can trigger internal "integerOneIndex" errors when
#   setting layout/threshold/appender later.
#
# Strategy:
#   - If indices exist, delete indices 2..n, then "re-initialize" index 1 by
#     setting its appender/formatter/layout/threshold.
#   - If no indices exist, we will create index 1 by calling log_appender()
#     first in logger.setup().
.logger_reset_namespace <- function() {
  if (!requireNamespace("logger", quietly = TRUE)) {
    stop("Package 'logger' must be installed to use MazamaCoreUtils logging.", call. = FALSE)
  }

  ns <- .MAZAMA_LOG_NAMESPACE
  n <- logger::log_indices(namespace = ns)

  if (is.na(n) || n <= 1L) {
    return(invisible(NULL))
  }

  # Delete in reverse order (but keep index 1).
  for (idx in seq.int(from = n, to = 2L, by = -1L)) {
    logger::delete_logger_index(namespace = ns, index = idx)
  }

  invisible(NULL)
}

# Quick test if logging has been initialized (MazamaCoreUtils view of the world)
.stopIfNotInitilized <- function() {
  if (!isTRUE(getOption("MazamaCoreUtils.logger.initialized", FALSE))) {
    stop(
      "You must initialize with 'logger.setup()' before issuing logger statements.",
      call. = FALSE
    )
  }
}

# ------------------------------------------------------------------------------
# Internal layout
# ------------------------------------------------------------------------------

.mazama_layout <- function(level, msg, namespace, index, ...) {

  # --------------------------------------------------------------------------
  # Determine level name
  # --------------------------------------------------------------------------

  # Preferred: loglevel objects carry the name as an attribute
  lvl <- attr(level, "level", exact = TRUE)

  # Fallback: bare integer (defensive, but rare)
  if (is.null(lvl)) {
    lvl <- switch(
      as.character(level),
      "0"   = "OFF",
      "100" = "FATAL",
      "200" = "ERROR",
      "300" = "WARN",
      "350" = "SUCCESS",
      "400" = "INFO",
      "500" = "DEBUG",
      "600" = "TRACE",
      as.character(level)
    )
  }

  # --------------------------------------------------------------------------
  # Format fields
  # --------------------------------------------------------------------------

  # Pad level to fixed width so timestamps align
  lvl <- sprintf("%-5s", toupper(lvl))

  # Timestamp in UTC with timezone
  ts <- format(
    as.POSIXct(Sys.time(), tz = "UTC"),
    "%Y-%m-%d %H:%M:%S %Z"
  )

  sprintf("%s [%s] %s", lvl, ts, msg)
}

# ------------------------------------------------------------------------------
# Public API
# ------------------------------------------------------------------------------

#' Set up Python-style logging
#'
#' Configure level-specific log files using the package logging API.
#'
#' Logging is built on top of the [logger](https://daroczig.github.io/logger/)
#' package while retaining the historical MazamaCoreUtils logging interface.
#'
#' Separate log files can be created for different log levels so that, for
#' example, an `errorLog` contains only `ERROR` and `FATAL` messages while a
#' `debugLog` contains `DEBUG` messages as well as all higher-severity messages.
#'
#' Any log file argument left as `NULL` is disabled and no file will be created
#' for that level.
#'
#' After initialization, logging statements can be generated with:
#' `logger.trace()`, `logger.debug()`, `logger.info()`,
#' `logger.warn()`, `logger.error()`, and `logger.fatal()`.
#'
#' @param traceLog File path receiving `TRACE` messages.
#' @param debugLog File path receiving `DEBUG` messages.
#' @param infoLog File path receiving `INFO` messages.
#' @param warnLog File path receiving `WARN` messages.
#' @param errorLog File path receiving `ERROR` messages.
#' @param fatalLog File path receiving `FATAL` messages.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @details
#' Log messages are formatted with:
#'
#' \preformatted{
#' LEVEL [YYYY-MM-DD HH:MM:SS UTC] message
#' }
#'
#' Console logging is enabled by default only for `FATAL` messages. Use
#' [logger.setLevel()] to display additional log messages in the console.
#'
#' @note
#' All functionality is implemented with the excellent \pkg{logger} package.
#'
#' @name logger.setup
#'
#' @importFrom logger log_appender log_layout log_threshold
#' @importFrom logger log_formatter
#' @importFrom logger appender_console appender_file appender_tee appender_void
#' @importFrom logger layout_simple
#' @importFrom logger formatter_sprintf
#' @export
#'
#' @examples
#' \dontrun{
#' # Create three log files
#' logger.setup(
#'   debugLog = "debug.log",
#'   infoLog = "info.log",
#'   errorLog = "error.log"
#' )
#'
#' # Generate log messages
#' logger.trace("trace statement #%d", 1)
#' logger.debug("debug statement")
#' logger.info("info statement %s %s", "with", "arguments")
#' logger.warn("warn statement: %s", "about to try something risky")
#'
#' result <- try(1 / "a", silent = TRUE)
#' logger.error("error message: %s", geterrmessage())
#' logger.fatal("fatal statement: %s", "THE END")
#'
#' cat(readLines("debug.log"), sep = "\n")
#' cat(readLines("info.log"), sep = "\n")
#' cat(readLines("error.log"), sep = "\n")
#' }
#'
#' @seealso
#' [logger.trace()], [logger.debug()], [logger.info()],
#' [logger.warn()], [logger.error()], [logger.fatal()]
#'
logger.setup <- function(
    traceLog = NULL,
    debugLog = NULL,
    infoLog  = NULL,
    warnLog  = NULL,
    errorLog = NULL,
    fatalLog = NULL
) {
  if (!requireNamespace("logger", quietly = TRUE)) {
    stop("Package 'logger' must be installed to use MazamaCoreUtils logging.", call. = FALSE)
  }

  ns <- .MAZAMA_LOG_NAMESPACE

  # Make setup idempotent (but do not delete index 1 down to zero indices).
  .logger_reset_namespace()

  # Ensure index 1 exists by setting an appender FIRST.
  # (logger will create the config slot when we set an appender.)
  if (is.null(fatalLog)) {
    # IMPORTANT: appender_console is a function; do NOT call it.
    logger::log_appender(logger::appender_console, namespace = ns, index = 1L)
  } else {
    if (file.exists(fatalLog)) file.remove(fatalLog)
    # appender_tee() is a factory (needs file), so DO call it.
    logger::log_appender(logger::appender_tee(fatalLog), namespace = ns, index = 1L)
  }

  # Use formatter_sprintf so existing "printf-style" calls continue to work.
  logger::log_formatter(logger::formatter_sprintf, namespace = ns, index = 1L)

  # layout_simple is a layout function; do NOT call it.
  logger::log_layout(.mazama_layout, namespace = ns, index = 1L)

  # By default, the console receives only FATAL messages.
  logger::log_threshold(.logger_map_level(FATAL), namespace = ns, index = 1L)

  # Helper for per-level file appenders (indices 2..6).
  # NOTE: Must set appender first (creates index config slot), then formatter/layout/threshold.
  set_file_logger <- function(index, level_const, path) {

    if (is.null(path)) {
      # Prefer built-in void appender; consistent with {logger}.
      logger::log_appender(logger::appender_void, namespace = ns, index = index)
    } else {
      if (file.exists(path)) file.remove(path)
      logger::log_appender(logger::appender_file(path), namespace = ns, index = index)
    }

    logger::log_formatter(logger::formatter_sprintf, namespace = ns, index = index)
    logger::log_layout(.mazama_layout, namespace = ns, index = index)
    logger::log_threshold(.logger_map_level(level_const), namespace = ns, index = index)

    invisible(NULL)
  }

  # Indices are fixed so downstream behavior is stable and easy to reason about.
  set_file_logger(index = 2L, level_const = TRACE, path = traceLog)
  set_file_logger(index = 3L, level_const = DEBUG, path = debugLog)
  set_file_logger(index = 4L, level_const = INFO,  path = infoLog)
  set_file_logger(index = 5L, level_const = WARN,  path = warnLog)
  set_file_logger(index = 6L, level_const = ERROR, path = errorLog)

  options("MazamaCoreUtils.logger.initialized" = TRUE)

  invisible(NULL)
}

#' Check whether logging has been initialized
#'
#' Determine whether [logger.setup()] has already been called.
#'
#' This function is useful in package code that conditionally emits log
#' statements only when logging has been configured.
#'
#' @return
#' Logical scalar indicating whether logging has been initialized.
#'
#' @name logger.isInitialized
#' @export
#'
#' @examples
#' \dontrun{
#' logger.isInitialized()
#'
#' logger.setup()
#'
#' logger.isInitialized()
#' }
#'
#' @seealso
#' [logger.setup()]
#'
logger.isInitialized <- function() {
  isTRUE(getOption("MazamaCoreUtils.logger.initialized", FALSE))
}

#' Set console log level
#'
#' Set the minimum log level displayed in the console.
#'
#' By default, only `FATAL` messages are displayed in the console. This
#' function allows users to display additional log messages interactively.
#'
#' Available log levels are:
#'
#' \preformatted{
#' TRACE
#' DEBUG
#' INFO
#' WARN
#' ERROR
#' FATAL
#' }
#'
#' @param level Logging threshold level.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @note
#' All functionality is implemented with the excellent \pkg{logger} package.
#'
#' @name logger.setLevel
#' @importFrom logger log_threshold
#' @export
#'
#' @examples
#' \dontrun{
#' # Enable console logging
#' logger.setup()
#'
#' # Show DEBUG and higher messages in the console
#' logger.setLevel(DEBUG)
#' }
#'
#' @seealso
#' [logger.setup()]
logger.setLevel <- function(level) {
  if (!logger.isInitialized()) {
    logger.setup()
  }

  lvl <- .logger_map_level(level)

  # Console is always index 1 in our namespace.
  invisible(logger::log_threshold(lvl, namespace = .MAZAMA_LOG_NAMESPACE, index = 1L))
}

# ------------------------------------------------------------------------------
# Logging functions (retain MazamaCoreUtils API)
# ------------------------------------------------------------------------------

#' Python-style logging statements
#'
#' Emit a `TRACE` level log message.
#'
#' Logging must first be initialized with [logger.setup()].
#'
#' @name logger.trace
#' @param msg Message with optional format strings.
#' @param ... Additional arguments passed to `sprintf()` formatting.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [logger.setup()]
#'
#' @export
#' @importFrom logger log_trace
logger.trace <- function(msg, ...) {
  .stopIfNotInitilized()
  logger::log_trace(msg, ..., namespace = .MAZAMA_LOG_NAMESPACE)
}

#' Python-style logging statements
#'
#' Emit a `DEBUG` level log message.
#'
#' Logging must first be initialized with [logger.setup()].
#'
#' @name logger.debug
#' @param msg Message with optional format strings.
#' @param ... Additional arguments passed to `sprintf()` formatting.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [logger.setup()]
#'
#' @export
#' @importFrom logger log_debug
#'
logger.debug <- function(msg, ...) {
  .stopIfNotInitilized()
  logger::log_debug(msg, ..., namespace = .MAZAMA_LOG_NAMESPACE)
}

#' Python-style logging statements
#'
#' Emit an `INFO` level log message.
#'
#' Logging must first be initialized with [logger.setup()].
#'
#' @name logger.info
#' @param msg Message with optional format strings.
#' @param ... Additional arguments passed to `sprintf()` formatting.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [logger.setup()]
#'
#' @export
#' @importFrom logger log_info
#'
logger.info <- function(msg, ...) {
  .stopIfNotInitilized()
  logger::log_info(msg, ..., namespace = .MAZAMA_LOG_NAMESPACE)
}

#' Python-style logging statements
#'
#' Emit a `WARN` level log message.
#'
#' Logging must first be initialized with [logger.setup()].
#'
#' @name logger.warn
#' @param msg Message with optional format strings.
#' @param ... Additional arguments passed to `sprintf()` formatting.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [logger.setup()]
#'
#' @export
#' @importFrom logger log_warn
#'
logger.warn <- function(msg, ...) {
  .stopIfNotInitilized()
  logger::log_warn(msg, ..., namespace = .MAZAMA_LOG_NAMESPACE)
}

#' Python-style logging statements
#'
#' Emit an `ERROR` level log message.
#'
#' Logging must first be initialized with [logger.setup()].
#'
#' @name logger.error
#' @param msg Message with optional format strings.
#' @param ... Additional arguments passed to `sprintf()` formatting.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [logger.setup()]
#'
#' @export
#' @importFrom logger log_error
#'
logger.error <- function(msg, ...) {
  .stopIfNotInitilized()
  logger::log_error(msg, ..., namespace = .MAZAMA_LOG_NAMESPACE)
}

#' Python-style logging statements
#'
#' Emit a `FATAL` level log message.
#'
#' Logging must first be initialized with [logger.setup()].
#'
#' @name logger.fatal
#' @param msg Message with optional format strings.
#' @param ... Additional arguments passed to `sprintf()` formatting.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @seealso
#' [logger.setup()]
#'
#' @export
#' @importFrom logger log_fatal
#'
logger.fatal <- function(msg, ...) {
  .stopIfNotInitilized()
  logger::log_fatal(msg, ..., namespace = .MAZAMA_LOG_NAMESPACE)
}

# ------------------------------------------------------------------------------
# Constants (retain existing exported API)
# ------------------------------------------------------------------------------

# Verbatim values from constants (legacy API). These are *not* required to match
# logger's internal numeric values; we map by name (see .logger_map_level()).

#' Log levels
#'
#' Logging level constants used by the MazamaCoreUtils logging API.
#'
#' Available log levels include:
#'
#' \preformatted{
#' FATAL
#' ERROR
#' WARN
#' INFO
#' DEBUG
#' TRACE
#' }
#'
#' These constants are retained for backwards compatibility with the original
#' MazamaCoreUtils logging system.
#'
#' @docType data
#' @name logLevels
#' @aliases FATAL ERROR WARN INFO DEBUG TRACE
#' @export
#'
FATAL <- 1L
names(FATAL) <- "FATAL"
#' @export
ERROR <- 2L
names(ERROR) <- "ERROR"
#' @export
WARN <- 4L
names(WARN) <- "WARN"
#' @export
INFO <- 6L
names(INFO) <- "INFO"
#' @export
DEBUG <- 8L
names(DEBUG) <- "DEBUG"
#' @export
TRACE <- 9L
names(TRACE) <- "TRACE"

test_that("initializeLogging() stops when logDir is invalid", {

  expect_error(initializeLogging(), "argument 'logDir' must not be NULL.")

})

test_that("initializeLogging() creates files", {

  logDir <- tempdir()
  initializeLogging(logDir)
  logger.trace("TRACE message")
  logger.debug("DEBUG message")
  logger.info("INFO message")
  logger.error("ERROR message")

  expect_true(
    all(file.exists(
      file.path(logDir, "TRACE.log"),
      file.path(logDir, "DEBUG.log"),
      file.path(logDir, "INFO.log"),
      file.path(logDir, "ERROR.log")
    ))
  )

  file.remove(
    file.path(logDir, "TRACE.log"),
    file.path(logDir, "DEBUG.log"),
    file.path(logDir, "INFO.log"),
    file.path(logDir, "ERROR.log")
  )

})

test_that("filePrefix is used", {

  logDir <- tempdir()
  initializeLogging(logDir, filePrefix = "bop_")
  logger.trace("TRACE message")
  logger.debug("DEBUG message")
  logger.info("INFO message")
  logger.error("ERROR message")

  expect_true(
    all(file.exists(
      file.path(logDir, "bop_TRACE.log"),
      file.path(logDir, "bop_DEBUG.log"),
      file.path(logDir, "bop_INFO.log"),
      file.path(logDir, "bop_ERROR.log")
    ))
  )

  file.remove(
    file.path(logDir, "bop_TRACE.log"),
    file.path(logDir, "bop_DEBUG.log"),
    file.path(logDir, "bop_INFO.log"),
    file.path(logDir, "bop_ERROR.log")
  )

})

# tests/testthat/test-logging.R

test_that("logging API functions exist", {
  expect_true(exists("logger.setup"))
  expect_true(exists("logger.isInitialized"))
  expect_true(exists("logger.setLevel"))
  expect_true(exists("logger.trace"))
  expect_true(exists("logger.debug"))
  expect_true(exists("logger.info"))
  expect_true(exists("logger.warn"))
  expect_true(exists("logger.error"))

  expect_true(is.function(logger.setup))
  expect_true(is.function(logger.isInitialized))
  expect_true(is.function(logger.setLevel))
})

test_that("logger calls error before initialization", {
  options("MazamaCoreUtils.logger.initialized" = FALSE)

  expect_error(
    logger.info("test")
  )
})

test_that("logger.isInitialized reflects setup state", {
  options("MazamaCoreUtils.logger.initialized" = FALSE)
  expect_false(logger.isInitialized())

  logger.setup()
  expect_true(logger.isInitialized())
})

test_that("logger.setup is idempotent and overwrites configured log files", {
  tf <- tempfile(fileext = ".log")

  logger.setup(traceLog = tf)
  logger.setLevel(TRACE)
  logger.trace("first")

  expect_true(file.exists(tf))
  expect_true(length(readLines(tf, warn = FALSE)) >= 1)

  # Run setup again: should remove and recreate the file
  logger.setup(traceLog = tf)
  logger.setLevel(TRACE)
  logger.trace("second")

  lines <- readLines(tf, warn = FALSE)
  expect_true(length(lines) >= 1)
  expect_false(any(grepl("first", lines, fixed = TRUE)))
  expect_true(any(grepl("second", lines, fixed = TRUE)))
})

test_that("logger.setLevel affects console threshold but does not break file logging", {
  tf_debug <- tempfile(fileext = ".log")

  logger.setup(debugLog = tf_debug)
  logger.setLevel(ERROR)   # console should be quiet for DEBUG

  logger.debug("dmsg")

  expect_true(any(grepl("dmsg", readLines(tf_debug, warn = FALSE), fixed = TRUE)))
})

test_that("log lines are properly formatted", {

  tf <- tempfile(fileext = ".log")

  # Single TRACE-level file captures all messages
  logger.setup(
    traceLog = tf
  )
  logger.setLevel(TRACE)

  logger.error("test")
  logger.warn("test")
  logger.info("test")
  logger.debug("test")
  logger.trace("test")

  lines <- readLines(tf, warn = FALSE)

  # We expect at least these 5 messages
  expect_true(length(lines) >= 5)
  tail5 <- tail(lines, 5)

  # "<LEVEL> [YYYY-MM-DD HH:MM:SS UTC] test"
  pattern <- "\\[[0-9-]{10} [0-9:]{8} UTC\\] test$"

  expect_match(tail5[1], paste0("^ERROR ", pattern))
  expect_match(tail5[2], paste0("^WARN  ", pattern))  # padded
  expect_match(tail5[3], paste0("^INFO  ", pattern))  # padded
  expect_match(tail5[4], paste0("^DEBUG ", pattern))
  expect_match(tail5[5], paste0("^TRACE ", pattern))

})

#' Load R data from a URL or local file
#'
#' Load a pre-generated R binary data file from either a local directory or a
#' remote URL.
#'
#' This function is intended for use by package-level `*_load()` helper
#' functions. It allows locally cached data files to be used when available,
#' avoiding unnecessary internet access.
#'
#' If both `dataDir` and `dataUrl` are provided, `priority` determines which
#' source is tried first. If loading from the first source fails, the other
#' source is used as a fallback.
#'
#' @param filename Name of the `.rda` file to load.
#' @param dataUrl Remote URL directory containing data files.
#' @param dataDir Local directory containing data files.
#' @param priority First data source to try when both `dataDir` and `dataUrl`
#'   are supplied.
#'
#' @return
#' Object loaded from the `.rda` file.
#'
#' @examples
#' \dontrun{
#' filename <- "USCensusStates_02.rda"
#' dataDir <- "~/Data/Spatial"
#' dataUrl <- "http://data.mazamascience.com/MazamaSpatialUtils/Spatial_0.8"
#'
#' # Load local file
#' USCensusStates <- loadDataFile(filename, dataDir = dataDir)
#'
#' # Load remote file
#' USCensusStates <- loadDataFile(filename, dataUrl = dataUrl)
#'
#' # Load local file with remote file as backup
#' USCensusStates <- loadDataFile(
#'   filename,
#'   dataDir = dataDir,
#'   dataUrl = dataUrl,
#'   priority = "dataDir"
#' )
#'
#' # Load remote file with local file as backup
#' USCensusStates <- loadDataFile(
#'   filename,
#'   dataDir = dataDir,
#'   dataUrl = dataUrl,
#'   priority = "dataUrl"
#' )
#' }
#'
#' @export
#'
loadDataFile <- function(
    filename = NULL,
    dataUrl = NULL,
    dataDir = NULL,
    priority = c("dataDir", "dataUrl")
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(filename)

  if ( is.null(dataUrl) && is.null(dataDir) ) {
    stop("either 'dataUrl' or 'dataDir' must be specified")
  }

  priority <- match.arg(priority)

  # ----- Load the data --------------------------------------------------------

  # * dataDir priority -----

  if ( !is.null(dataDir) && (is.null(dataUrl) || priority == "dataDir") ) {

    # Don't stop yet
    result <- try({
      loadedData <- .loadFromDir(filename, dataDir)
    }, silent = TRUE)

    if ( "try-error" %in% class(result) ) {
      if ( is.null(dataUrl) ) {
        stop(toString(result), call. = FALSE)
      } else {
        # Load form URL or stop
        try({
          loadedData <- .loadFromUrl(filename, dataUrl)
        }, silent = TRUE) %>%
          stopOnError("data file could not be loaded from dataDir or dataUrl")
      }
    }

  }

  # * dataUrl priority -----

  if ( !is.null(dataUrl) && (is.null(dataDir) || priority == "dataUrl") ) {

    # Don't stop yet
    result <- try({
      loadedData <- .loadFromUrl(filename, dataUrl)
    }, silent = TRUE)

    if ( "try-error" %in% class(result) ) {
      if ( is.null(dataDir) ) {
        stop(toString(result), call. = FALSE)
      } else {
        # Load from directory or stop
        try({
          loadedData <- .loadFromDir(filename, dataDir)
        }, silent = TRUE) %>%
          stopOnError("data file could not be loaded from dataDir or dataUrl")
      }
    }

  }

  # ----- Return ---------------------------------------------------------------

  return(loadedData)

}

# ===== INTERNAL FUNCTIONS =====================================================

.loadFromDir <- function(filename, dataDir) {

  dataDir <- path.expand(dataDir)

  if ( !dir.exists(dataDir) )
    stop(sprintf("dataDir '%s' does not exist.", dataDir), call. = FALSE)

  filepath <- file.path(dataDir, filename)

  result <- try({
    suppressWarnings({
      objectName <- load(filepath)
    })
  }, silent = TRUE)

  if ( "try-error" %in% class(result) ) {
    stop(sprintf("data file could not be loaded from: %s", filepath), call. = FALSE)
  }

  # No error
  loadedData <- get(objectName)
  return(loadedData)

}


.loadFromUrl <- function(filename, dataUrl) {

  filepath <- file.path(dataUrl, filename)

  # Define a 'connection' object so we can close it no matter what happens
  conn <- url(filepath)
  on.exit(close(conn), add = TRUE)
  result <- try({
    suppressWarnings({
      objectName <- load(conn)
    })
  }, silent = TRUE)

  if ( "try-error" %in% class(result) ) {
    stop(sprintf("data file could not be loaded from: %s", filepath), call. = FALSE)
  }

  # No error
  loadedData <- get(objectName)
  return(loadedData)

}

#' Lint function calls for required named arguments
#'
#' Parse R source code and identify calls to selected functions that are missing
#' required named arguments.
#'
#' Rules are supplied as a named list where each name is a function to check and
#' each value is a character vector of required argument names. A function call
#' passes when all required arguments are supplied by name.
#'
#' This linter only checks whether arguments are named in the call. It does not
#' evaluate code, inspect argument values, or detect unnamed positional
#' arguments.
#'
#' @param filePath Path to a single R source file.
#' @param dirPath Path to a directory containing R source files.
#' @param rules Named list of linting rules. Each list name is a function name
#'   and each value is a character vector of required named arguments.
#' @param fullPath Logical specifying whether returned file paths should be
#'   absolute paths. If `FALSE`, only base file names are returned.
#'
#' @return
#' A tibble describing matching function calls, with columns:
#'
#' \describe{
#'   \item{file}{Source file path or file name.}
#'   \item{line_number}{Line number where the function call begins.}
#'   \item{column_number}{Column number where the function call begins.}
#'   \item{function_name}{Name of the function being checked.}
#'   \item{named_args}{List column containing named arguments used in the call.}
#'   \item{includes_required}{Logical indicating whether all required named
#'     arguments were supplied.}
#' }
#'
#' @section Limitations:
#' This linter only detects named arguments. For example, `foo(x = bar, "baz")`
#' is treated as specifying the named argument `x`, but the value `bar` and the
#' unnamed argument `"baz"` are not inspected.
#'
#' @name lintFunctionArgs
#' @aliases lintFunctionArgs_file lintFunctionArgs_dir
#'
#' @examples
#' \dontrun{
#' rules <- list(
#'   fn_one = "x",
#'   fn_two = c("foo", "bar")
#' )
#'
#' lintFunctionArgs_file(
#'   filePath = "local_test/timezone_lint_test_script.R",
#'   rules = rules
#' )
#'
#' lintFunctionArgs_dir(
#'   dirPath = "./R",
#'   rules = MazamaCoreUtils::timezoneLintRules
#' )
#' }
NULL
NULL

#' @rdname lintFunctionArgs
#' @export
lintFunctionArgs_file <- function(
  filePath = NULL,
  rules = NULL,
  fullPath = FALSE
) {

  # Validate input ------------------------------------------------------------

  stopIfNull(filePath)
  stopIfNull(rules)

  if ( !is.list(rules) || is.null(names(rules)) ) {
    stop("rules must be a named list")
  }

  if ( !is.character(filePath) || length(filePath) != 1 ) {
    stop("filePath must be a length 1 character vector")
  }

  normFilePath <- normalizePath(filePath)

  if ( !utils::file_test("-f", normFilePath) ) {
    stop("filePath must point to a file, not a directory")
  }


  # Parse file ----------------------------------------------------------------

  parsedData <-
    normFilePath %>%
    parse(keep.source = TRUE) %>%
    utils::getParseData() %>%
    tibble::as_tibble()


  # Collect functions and arguments -----------------------------------------

  # Given IDs as names, this vector outputs the IDs' parent IDs
  lookupParent <-
    parsedData %>%
    dplyr::select(.data$id, .data$parent) %>%
    tibble::deframe()

  # Group function arguments by which function they belong to
  functionArgs <-
    parsedData %>%
    dplyr::filter(.data$token == "SYMBOL_SUB") %>%
    dplyr::group_by(.data$parent) %>%
    dplyr::summarise(named_args = list(.data$text)) %>%
    dplyr::rename(id = .data$parent)

  # Pair function calls with their arguments
  functionCalls <-
    parsedData %>%
    dplyr::mutate(lookup_pid = lookupParent[as.character(.data$parent)]) %>%
    dplyr::filter(.data$token == "SYMBOL_FUNCTION_CALL") %>%
    dplyr::select(
      line_number = .data$line1,
      column_number = .data$col1,
      function_name = .data$text,
      id = .data$lookup_pid
    ) %>%
    dplyr::left_join(functionArgs, by = "id") %>%
    dplyr::select(-.data$id)


  # Check function arguments ------------------------------------------------

  fileString <- if ( fullPath ) normFilePath else basename(normFilePath)

  results <-
    functionCalls %>%
    dplyr::filter(.data$function_name %in% names(rules)) %>%
    dplyr::mutate(
      includes_required = purrr::map2_lgl(
        .data$named_args,
        .data$function_name,
        ~all(rules[[.y]] %in% .x)
      ),
      file = fileString
    ) %>%
    dplyr::select(.data$file, dplyr::everything())

  return(results)

}


#' @rdname lintFunctionArgs
#' @export
lintFunctionArgs_dir <- function(
  dirPath = "./R",
  rules = NULL,
  fullPath = FALSE
) {

  # Validate input -------------------------------------------------------------

  stopIfNull(rules)

  if ( !is.list(rules) || is.null(names(rules)) )
    stop("rules must be a named list")

  if ( !is.character(dirPath) || length(dirPath) != 1 )
    stop("dirPath must be a length 1 character vector")

  normDirPath <- normalizePath(dirPath)

  if ( !utils::file_test("-d", normDirPath) )
    stop("filePath must point to a directory, not a file")


  # Lint files -----------------------------------------------------------------

  results <- normDirPath %>%
    list.files(pattern = "\\.R$", full.names = TRUE, recursive = TRUE) %>%
    purrr::map_dfr(lintFunctionArgs_file, rules, fullPath)

  return(results)

}

#' Timezone linting rules
#'
#' Rules used by [lintFunctionArgs_file()] and [lintFunctionArgs_dir()] to find
#' date/time function calls that should explicitly specify timezone arguments.
#'
#' Each list name is a function to check. Each value is the required named
#' timezone-related argument for that function.
#'
#' Entries with `"DEPRECATED"` are used to flag functions that should generally
#' be avoided in package code because they depend on the local system clock or
#' timezone.
#'
#' @format
#' A named list of function/argument pairs.
#'
#' @examples
#' str(timezoneLintRules)
#'
#' @docType data
#' @name timezoneLintRules
#' @export
#'
timezoneLintRules <- list(
  # base functions
  "as.Date" = "tz",
  "as.POSIXct" = "tz",
  "as.POSIXlt" = "tz",
  "ISOdate" = "tz",
  "ISOdatetime" = "tz",
  "strftime" = "tz",
  "strptime" = "tz",
  "Sys.Date" = "DEPRECATED", # Please don't use this function!
  "Sys.time" = "DEPRECATED", # Please don't use this function!
  # lubridate functions
  "as_datetime" = "tz",
  "date_decimal" = "tz",
  "dmy" = "tz",
  "dmy_h" = "tz",
  "dmy_hm" = "tz",
  "dmy_hms" = "tz",
  "dym" = "tz",
  "fast_strptime" = "tz",
  "force_tz" = "tzone",
  "force_tzs" = "tzone_out",
  "interval" = "tzone",
  "local_time" = "tz",
  "make_datetime" = "tz",
  "mdy" = "tz",
  "mdy_h" = "tz",
  "mdy_hm" = "tz",
  "mdy_hms" = "tz",
  "my" = "tz",
  "myd" = "tz",
  "now" = "tzone",
  "parse_date_time" = "tz",
  "parse_date_time2" = "tz",
  "today" = "tzone",
  "with_tz" = "tzone",
  "ydm" = "tz",
  "ydm_h" = "tz",
  "ydm_hm" = "tz",
  "ydm_hms" = "tz",
  "ym" = "tz",
  "ymd" = "tz",
  "ymd_h" = "tz",
  "ymd_hm" = "tz",
  "ymd_hms" = "tz",
  "yq" = "tz",
  # MazamaCoreUtils functions
  "dateRange" = "timezone",
  "timeRange" = "timezone",
  "timeStamp" = "timezone",
  "parseDatetime" = "timezone"
)

#' Extract links from an HTML page
#'
#' Parse an HTML page and return all `<a href="...">...</a>` links as a data
#' frame.
#'
#' The returned data frame contains the human-readable link text in `linkName`
#' and the `href` value in `linkUrl`. This is useful for extracting links from
#' index pages, including web-accessible directories that list downloadable
#' files.
#'
#' Wrapper functions [html_getLinkNames()] and [html_getLinkUrls()] return the
#' corresponding columns as character vectors.
#'
#' @param url URL or local file path of an HTML page.
#' @param relative Logical specifying whether to return relative URLs. If
#'   `FALSE`, relative URLs are converted to absolute URLs using `url` as the
#'   base.
#'
#' @return
#' A tibble with `linkName` and `linkUrl` columns.
#'
#' @examples
#' \dontrun{
#'
#' # If you want to download lots of USCensus shapefiles
#' url <- "https://www2.census.gov/geo/tiger/GENZ2019/shp/"
#'
#' browseURL(url)
#'
#' dataLinks <- html_getLinks(url)
#'
#' dataLinks <-
#'   dataLinks %>%
#'   dplyr::filter(stringr::str_detect(linkName, "us_county"))
#'
#' head(dataLinks, 10)
#'
#' html_getLinkNames(url)
#' html_getLinkUrls(url, relative = FALSE)
#' }
#'
#' @name html_getLinks
#' @rdname html_getLinks
#' @export

html_getLinks <- function(
  url = NULL,
  relative = TRUE
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(url)

  if ( !is.logical(relative) )
    relative <- TRUE

  # ----- Extract the links ----------------------------------------------------

  try({

    urlAttributes <- xml2::read_html(url) %>% xml2::xml_find_all("//a[@href]")

    urlText <-
      urlAttributes %>%
      xml2::xml_text()

    urlLinks <-
      urlAttributes %>%
      xml2::xml_attr("href")

  }, silent = TRUE) %>%
  stopOnError()

  df <- dplyr::tibble(linkName = urlText, linkUrl = urlLinks)

  # ----- Filter URLs -------------------------------------------------

  df <-
    df %>%

    # Remove NA values
    dplyr::filter(!is.na(.data$linkUrl) & !is.na(.data$linkName)) %>%

    # Remove Apache indexing
    dplyr::filter(stringr::str_detect(.data$linkUrl, "^\\?C=.;O=.*", negate = TRUE)) %>%

    # Remove "Parent Directory"
    dplyr::filter(stringr::str_detect(.data$linkName, "Parent Directory", negate = TRUE)) %>%

    # Format URLs beginning with //
    dplyr::mutate(linkUrl = stringr::str_replace(.data$linkUrl, stringr::regex("^//"), ""))


  # ----- Expand relative URLs -------------------------------------------------

  if ( !relative ) {

    df <-
      df %>%
      dplyr::mutate(
        linkUrl = xml2::url_absolute(.data$linkUrl, base = url)
      )

  }

  # ----- Return ---------------------------------------------------------------

  return(df)

}

#' @rdname html_getLinks
#'
#' @return
#' `html_getLinkNames()` returns a character vector of link names.
#'
#' @export
html_getLinkNames <- function(
  url = NULL
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(url)

  # ----- Extract the link text ------------------------------------------------

  linkNames <-
    html_getLinks(url) %>%
    dplyr::pull("linkName")

  # ----- Return ---------------------------------------------------------------

  return(linkNames)

}

#' @rdname html_getLinks
#'
#' @return
#' `html_getLinkUrls()` returns a character vector of link URLs.
#'
#' @export
html_getLinkUrls <- function(
  url = NULL,
  relative = TRUE
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(url)

  # ----- Extract the link text ------------------------------------------------

  linkUrls <-
    html_getLinks(url, relative) %>%
    dplyr::pull("linkUrl")

  # ----- Return ---------------------------------------------------------------

  return(linkUrls)

}

#' Validate longitude and latitude vectors
#'
#' Validate longitude and latitude vectors to ensure they are numeric,
#' have matching lengths, and contain values within valid geographic bounds.
#'
#' Longitudes must fall between -180 and 180 degrees and latitudes must
#' fall between -90 and 90 degrees. If validation fails, an error is
#' generated.
#'
#' @param longitude Vector of longitudes in decimal degrees east.
#' @param latitude Vector of latitudes in decimal degrees north.
#' @param na.rm Logical specifying whether to remove `NA` values before
#'   validation.
#'
#' @return
#' Invisibly returns `TRUE` if validation succeeds.
#'
#' @examples
#' longitude <- c(-122.5, -122.4)
#' latitude <- c(47.5, 47.6)
#'
#' validateLonsLats(longitude, latitude)
#'
#' # Remove missing values before validation
#' validateLonsLats(
#'   c(-122.5, NA),
#'   c(47.5, NA),
#'   na.rm = TRUE
#' )
#'
#' \dontrun{
#' validateLonsLats(c(-200, 0), c(45, 46))
#' }
#'
#' @rdname validateLonsLats
#' @export
validateLonsLats <- function(
    longitude = NULL,
    latitude = NULL,
    na.rm = FALSE
) {

  # ----- Validate parameters --------------------------------------------------

  MazamaCoreUtils::stopIfNull(longitude)
  MazamaCoreUtils::stopIfNull(latitude)

  if ( !is.numeric(longitude) )
    stop("'longitude' must be numeric")

  if ( !is.numeric(latitude) )
    stop("'latitude' must be numeric")

  if ( length(longitude) != length(latitude) )
    stop("'longitude' and 'latitude' must have the same length")

  # Remove locations with NAs
  if ( na.rm ) {
    good_mask <- !is.na(longitude) & !is.na(latitude)
    longitude <- longitude[good_mask]
    latitude <- latitude[good_mask]
  }

  if ( anyNA(longitude) || any(longitude < -180) || any(longitude > 180 ))
    stop("all longitudes must be valid values between -180 and 180")

  if ( anyNA(latitude) || any(latitude < -90) || any(latitude > 90) )
    stop("all latitudes must be valid values between -90 and 90")

  return(invisible(TRUE))

}


#' Validate longitude and latitude values
#'
#' Validate a single longitude/latitude pair to ensure both values are
#' numeric scalars and fall within valid geographic bounds.
#'
#' Longitudes must fall between -180 and 180 degrees and latitudes must
#' fall between -90 and 90 degrees. If validation fails, an error is
#' generated.
#'
#' @param longitude Single longitude in decimal degrees east.
#' @param latitude Single latitude in decimal degrees north.
#'
#' @return
#' Invisibly returns `TRUE` if validation succeeds.
#'
#' @examples
#' validateLonLat(-122.5, 47.5)
#'
#' \dontrun{
#' validateLonLat(-200, 47.5)
#' validateLonLat(-122.5, NA)
#' }
#'
#' @rdname validateLonLat
#' @export
validateLonLat <- function(
    longitude = NULL,
    latitude = NULL
) {

  MazamaCoreUtils::stopIfNull(longitude)
  MazamaCoreUtils::stopIfNull(latitude)

  if ( !is.numeric(longitude) )
    stop("'longitude' must be numeric")

  if ( !is.numeric(latitude) )
    stop("'latitude' must be numeric")

  if ( length(longitude) > 1 || length(latitude) > 1 ) {
    stop("longitude and latitude must be single values")
  }

  if ( is.na(longitude) || longitude < -180 || longitude > 180 )
    stop("'longitude' must be a valid value between -180 and 180")

  if ( is.na(latitude) || latitude < -90 || latitude > 90 )
    stop("'latitude' must be a valid value between -90 and 90")

  return(invisible(TRUE))

}

test_that("validateLonLat() works", {

  expect_true(validateLonLat(-120.325278, 47.423333))

  expect_error(validateLonLat())
  expect_error(validateLonLat("abc"))
  expect_error(validateLonLat(-120.325278))
  expect_error(validateLonLat(-120.325278, "abc"))
  expect_error(validateLonLat(-120.325278, -91))
  expect_error(validateLonLat(-120.325278, 91))
  expect_error(validateLonLat(-181, 47.423333))
  expect_error(validateLonLat(181, 47.423333))

})

test_that("validateLonsLats() works", {

  lons <- 60:80
  lats <- 20:40

  expect_true(validateLonsLats(lons, lats))

  expect_error(validateLonsLats())
  expect_error(validateLonsLats("abc"))
  expect_error(validateLonsLats(lons))

  lons[1] <- NA
  expect_error(validateLonsLats(lons, lats))
  lons[1] <- -181
  expect_error(validateLonsLats(lons, lats))
  lons[1] <- 181
  expect_error(validateLonsLats(lons, lats))

  lons[1] <- 60
  lats[1] <- NA
  expect_error(validateLonsLats(lons, lats))
  lons[1] <- -91
  expect_error(validateLonsLats(lons, lats))
  lons[1] <- 91
  expect_error(validateLonsLats(lons, lats))

})

test_that("createLocationMask() validates input", {

  # Generate errors
  expect_error(createLocationMask())
  expect_error(createLocationMask("abc"))
  expect_error(createLocationMask(-120.325278))
  expect_error(createLocationMask(1:4, 1:5))
  expect_error(createLocationMask(1, 1, lonRange = c(-200, -120)))
  expect_error(createLocationMask(1, 1, latRange = c(-200, 500)))

  # Acceptable bad input
  expect_equal(createLocationMask(-120.325278, "abc"), FALSE)
  expect_equal(createLocationMask(-120.325278, NA), FALSE)

})

test_that("createLocationMask() works", {

  # removeZeroZero = TRUE
  expect_equal(
    createLocationMask(
      longitude = c(1,NA,1,0,1,200,  1,1,1,0,1,1),
      latitude = c(1,1,1,1,1,1,  1,NA,1,0,1,200)
    ),
    c(TRUE, FALSE, TRUE, TRUE, TRUE, FALSE,
      TRUE, FALSE, TRUE, FALSE, TRUE, FALSE)
  )

  # removeZeroZero = FALSE
  expect_equal(
    createLocationMask(
      longitude = c(1,NA,1,0,1,200,  1,1,1,0,1,1),
      latitude = c(1,1,1,1,1,1,  1,NA,1,0,1,200),
      removeZeroZero = FALSE
    ),
    c(TRUE, FALSE, TRUE, TRUE, TRUE, FALSE,
      TRUE, FALSE, TRUE, TRUE, TRUE, FALSE)
  )

})

#' Create a POSIXct time range
#'
#' Create an ordered two-element `POSIXct` time range from start and end
#' datetime values.
#'
#' Input values are converted with [parseDatetime()] using the required
#' `timezone` argument. The resulting start and end times are sorted so the
#' earlier time is always returned first.
#'
#' By default, both times are rounded down with [lubridate::floor_date()] using
#' the requested `unit`. Set `ceilingStart = TRUE` or `ceilingEnd = TRUE` to
#' round either endpoint up with [lubridate::ceiling_date()] instead.
#'
#' @param starttime Desired start datetime.
#' @param endtime Desired end datetime.
#' @param timezone Olson timezone used to interpret incoming datetimes.
#' @param unit Unit used for rounding. Passed to [lubridate::floor_date()] or
#'   [lubridate::ceiling_date()].
#' @param ceilingStart Logical specifying whether to round the start time up
#'   instead of down.
#' @param ceilingEnd Logical specifying whether to round the end time up instead
#'   of down.
#'
#' @return
#' Two-element `POSIXct` vector ordered from earliest to latest.
#'
#' @inheritSection dateRange POSIXct inputs
#'
#' @examples
#' timeRange(
#'   starttime = "2019-01-08 10:12:15",
#'   endtime = 20190109102030,
#'   timezone = "UTC"
#' )
#'
#' timeRange(
#'   starttime = "2019-01-08 10:12:15",
#'   endtime = "2019-01-09 10:20:30",
#'   timezone = "UTC",
#'   unit = "hour"
#' )
#'
#' @export
#'
timeRange <- function(
  starttime = NULL,
  endtime = NULL,
  timezone = NULL,
  unit = "sec",
  ceilingStart = FALSE,
  ceilingEnd = FALSE
) {

  # ----- Validate parameters --------------------------------------------------

  stopIfNull(starttime)
  stopIfNull(endtime)
  stopIfNull(timezone)
  setIfNull(unit, "sec")
  setIfNull(ceilingStart, FALSE)
  setIfNull(ceilingEnd, FALSE)

  if ( !timezone %in% base::OlsonNames() )
    stop(sprintf("'timezone = %s' is not found in OlsonNames()", timezone))

  # ----- Process datetimes ----------------------------------------------------

  # Guarantee conversion to POSIXct
  starttime <- parseDatetime(starttime, timezone = timezone)
  endtime <- parseDatetime(endtime, timezone = timezone)

  # Guarantee proper ordering
  timeRange <- sort(c(starttime, endtime))

  # Floor/Ceiling to nearest unit
  if ( ceilingStart ) {
    timeRange[1] <- lubridate::ceiling_date(timeRange[1], unit = unit)
  } else {
    timeRange[1] <- lubridate::floor_date(timeRange[1], unit = unit)
  }

  if ( ceilingEnd ) {
    timeRange[2] <- lubridate::ceiling_date(timeRange[2], unit = unit)
  } else {
    timeRange[2] <- lubridate::floor_date(timeRange[2], unit = unit)
  }

  # ----- Return ---------------------------------------------------------------

  return(timeRange)

}

test_that("timeRange() requires explicit timezone", {

  expect_error(timeRange(timezone = NULL))

})


test_that("Output is a two element POSIXct vector", {


  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 8, 20, 30, 40, tz = "America/Los_Angeles")

  tlim <- timeRange(starttime = startTime, endtime = endTime, timezone = "America/Los_Angeles")

  expect_length(tlim, 2)
  expect_s3_class(tlim, "POSIXct")

})


test_that("Output has the correct timezone", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 8, 20, 30, 40, tz = "America/Los_Angeles")

  tlim_LA <- timeRange(starttime = startTime, endtime = endTime, timezone = "America/Los_Angeles")
  tlim_UTC <- timeRange(starttime = startTime, endtime = endTime, timezone = "UTC")

  expect_identical(attributes(tlim_LA)$tzone, "America/Los_Angeles")
  expect_identical(attributes(tlim_UTC)$tzone, "UTC")

})


test_that("First element in output is less than the second", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 8, 20, 30, 40, tz = "America/Los_Angeles")

  tlim <- timeRange(starttime = startTime, endtime = endTime, timezone = "America/Los_Angeles")
  tlim_rev <- timeRange(starttime = endTime, endtime = startTime, timezone = "America/Los_Angeles")

  expect_lt(tlim[1], tlim[2])
  expect_lt(tlim_rev[1], tlim_rev[2])

})


test_that("Non-POSIXct inputs work as expected", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 8, 20, 30, 40, tz = "America/Los_Angeles")

  startTime_string <- strftime(startTime, format = "%Y-%m-%d %H:%M:%S", tz = "America/Los_Angeles")
  starTime_numeric <- strftime(startTime, format = "%Y%m%d%H%M%S", tz = "America/Los_Angeles") %>% as.numeric()

  endTime_string <- strftime(endTime, format = "%Y-%m-%d %H:%M:%S", tz = "America/Los_Angeles")
  endTime_numeric <- strftime(endTime, format = "%Y%m%d%H%M%S", tz = "America/Los_Angeles") %>% as.numeric()

  tlim_posixct <- timeRange(starttime = startTime, endtime = endTime, timezone = "America/Los_Angeles")
  tlim_string <- timeRange(starttime = startTime_string, endtime = endTime_string, timezone = "America/Los_Angeles")
  tlim_numeric <- timeRange(starttime = starTime_numeric, endtime = endTime_numeric, timezone = "America/Los_Angeles")

  expect_identical(tlim_posixct, tlim_string)
  expect_identical(tlim_posixct, tlim_numeric)

})


test_that("unit, ceilingStart and ceilingEnd arguments work", {

  startTime <- ISOdatetime(2019, 08, 01, 2, 15, 45, tz = "America/Los_Angeles")
  endTime <- ISOdatetime(2019, 08, 8, 20, 30, 40, tz = "America/Los_Angeles")

  startMin <- ISOdatetime(2019, 08, 01, 2, 15, 00, tz = "America/Los_Angeles")
  startHour <- ISOdatetime(2019, 08, 01, 2, 00, 00, tz = "America/Los_Angeles")
  startDay <- ISOdatetime(2019, 08, 01, 00, 00, 00, tz = "America/Los_Angeles")

  startMinCeil <- ISOdatetime(2019, 08, 01, 2, 16, 00, tz = "America/Los_Angeles")
  startHourCeil <- ISOdatetime(2019, 08, 01, 3, 00, 00, tz = "America/Los_Angeles")
  startDayCeil <- ISOdatetime(2019, 08, 02, 00, 00, 00, tz = "America/Los_Angeles")

  endMin <- ISOdatetime(2019, 08, 8, 20, 30, 00, tz = "America/Los_Angeles")
  endHour <- ISOdatetime(2019, 08, 8, 20, 00, 00, tz = "America/Los_Angeles")
  endDay <- ISOdatetime(2019, 08, 8, 00, 00, 00, tz = "America/Los_Angeles")

  endMinCeil <- ISOdatetime(2019, 08, 8, 20, 31, 00, tz = "America/Los_Angeles")
  endHourCeil <- ISOdatetime(2019, 08, 8, 21, 00, 00, tz = "America/Los_Angeles")
  endDayCeil <- ISOdatetime(2019, 08, 9, 00, 00, 00, tz = "America/Los_Angeles")

  tlim <- timeRange(startTime, endTime, timezone = "America/Los_Angeles")

  tlimMin <- timeRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "min")
  tlimHour <- timeRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "hour")
  tlimDay <- timeRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "day")

  tlimMinCeil <- timeRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "min", ceilingStart = TRUE, ceilingEnd = TRUE)
  tlimHourCeil <- timeRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "hour", ceilingStart = TRUE, ceilingEnd = TRUE)
  tlimDayCeil <- timeRange(startTime, endTime, timezone = "America/Los_Angeles", unit = "day", ceilingStart = TRUE, ceilingEnd = TRUE)

  expect_identical(tlim, c(startTime, endTime))
  expect_identical(tlimMin, c(startMin, endMin))
  expect_identical(tlimHour, c(startHour, endHour))
  expect_identical(tlimDay, c(startDay, endDay))
  expect_identical(tlimMinCeil, c(startMinCeil, endMinCeil))
  expect_identical(tlimHourCeil, c(startHourCeil, endHourCeil))
  expect_identical(tlimDayCeil, c(startDayCeil, endDayCeil))

})

#' Stop if an object is `NULL`
#'
#' Convenience function for validating that an object is not `NULL`.
#'
#' If `target` is not `NULL`, it is returned invisibly. If `target` is
#' `NULL`, the function stops with either a default or user-supplied
#' error message.
#'
#' This function is especially useful for validating required function
#' arguments or for guarding intermediate results in pipelines.
#'
#' @param target Object to test.
#' @param msg Optional error message to display if `target` is `NULL`.
#'   Must be a character string of length one.
#'
#' @return
#' Invisibly returns `target` when it is not `NULL`.
#'
#' @examples
#' # Return input invisibly if not NULL
#' x <- stopIfNull(5)
#' print(x)
#'
#' # Useful in pipelines
#' y <- 1:10
#' y_mean <-
#'   y %>%
#'   stopIfNull() %>%
#'   mean()
#'
#' \dontrun{
#' # Trigger the default error message
#' testVar <- NULL
#' stopIfNull(testVar)
#'
#' # Trigger a custom error message
#' stopIfNull(testVar, msg = "This is NULL")
#'
#' # Make a failing pipeline
#' z <- NULL
#' z_mean <-
#'   z %>%
#'   stopIfNull("This has failed.") %>%
#'   mean()
#' }
#'
#' @export
stopIfNull <- function(
    target,
    msg = NULL
) {

  # Return early if not NULL ---------------------------------------------------

  if ( !is.null(target) )
    return(invisible(target))


  # Build error message --------------------------------------------------------

  if ( is.null(msg) ) {

    msg <- paste0(
      "argument '",
      deparse(substitute(target)),
      "' must not be NULL."
    )

  } else if ( !is.character(msg) || length(msg) != 1 ) {

    stop("'msg' must be a character string of length one")

  }


  # Stop -----------------------------------------------------------------------

  stop(msg)

}

test_that("Non-null inputs are returned invisibly", {

  expect_invisible(stopIfNull(10))
  expect_invisible(stopIfNull("bar"))
  expect_invisible(stopIfNull(mtcars))

  expect_identical(stopIfNull(5), 5)
  expect_identical(stopIfNull("foo"), "foo")
  expect_identical(stopIfNull(mtcars), mtcars)

})


test_that("Error messages work as expected", {

  customMessage <- "This is a message!"

  expect_error(stopIfNull(NULL))
  expect_error(stopIfNull(NULL, msg = customMessage), customMessage)

})

#' Stop on try-error
#'
#' Generate a consistent error message from the result of a `try()` block.
#'
#' This function is intended for production code where potentially fragile
#' operations are wrapped in `try(..., silent = TRUE)`. If `result` inherits
#' from `"try-error"`, a cleaned and optionally customized error message is
#' generated and passed to [stop()].
#'
#' If `result` is not a `"try-error"`, the function returns `NULL`.
#'
#' @param result Return value from a `try()` block.
#' @param err_msg Optional custom error message.
#' @param prefix Optional text to prepend to the error message.
#' @param maxLength Maximum allowed error message length before truncation.
#' @param truncatedLength Length of the truncated error message.
#' @param call. Logical indicating whether the call should be included in the
#'   error message. Passed to [stop()].
#'
#' @return
#' Returns `NULL` if `result` is not a `"try-error"`; otherwise stops with an
#' error.
#'
#' @note
#' If logging has been initialized, the final error message is logged with
#' [logger.error()] before calling [stop()].
#'
#' @examples
#' \dontrun{
#' myFunc <- function(x) {
#'   log(x)
#' }
#'
#' result <- try({
#'   myFunc("ten")
#' }, silent = TRUE)
#'
#' stopOnError(result)
#'
#' try({
#'   myFunc("ten")
#' }, silent = TRUE) %>%
#'   stopOnError(err_msg = "Unable to process user input")
#'
#' try({
#'   myFunc("ten")
#' }, silent = TRUE) %>%
#'   stopOnError(
#'     prefix = "USER_INPUT_ERROR",
#'     maxLength = 40,
#'     truncatedLength = 32
#'   )
#' }
#'
#' @name stopOnError
#' @export

stopOnError <- function(
  result,
  err_msg = "",
  prefix = "",
  maxLength = 500,
  truncatedLength = 120,
  call. = FALSE
) {

  if ( "try-error" %in% class(result) ) {

    # Use passed in message or cleaned up version from geterrmessage()
    err_msg <- ifelse(err_msg == "", geterrmessage(), err_msg)

    err_msg <-
      err_msg %>%
      stringr::str_replace("Error : ", "") %>%
      stringr::str_replace("Error: ", "") %>%
      stringr::str_trim()

    if ( prefix != "" )
      err_msg <- paste(prefix, err_msg)

    if ( stringr::str_length(err_msg) > maxLength )
      err_msg <- paste(stringr::str_sub(err_msg, end = truncatedLength), "...")

    if ( logger.isInitialized() )
      logger.error(err_msg)

    stop(err_msg, call. = call.)

  }

}

#' Initialize standard log files
#'
#' Create a standard set of MazamaCoreUtils log files.
#'
#' This convenience function creates or validates a log directory, archives any
#' existing standard log files by appending a UTC timestamp, and then initializes
#' logging with [logger.setup()].
#'
#' Standard log files include:
#'
#' \preformatted{
#' TRACE.log
#' DEBUG.log
#' INFO.log
#' WARN.log
#' ERROR.log
#' }
#'
#' When `filePrefix` is supplied, it is prepended to each log file name.
#'
#' @param logDir Directory in which to write log files.
#' @param filePrefix Character string prepended to log file names.
#' @param createDir Logical specifying whether to create `logDir` if it does
#'   not already exist.
#'
#' @return
#' No return value. Called for side effects.
#'
#' @name initializeLogging
#' @export
#'
#' @seealso
#' [logger.setup()]
#'
initializeLogging <- function(
  logDir = NULL,
  filePrefix = "",
  createDir = TRUE
) {

  # ----- Validate parameters --------------------------------------------------

  stopIfNull(logDir)
  filePrefix <- setIfNull(filePrefix, "")
  createDir <- setIfNull(createDir, TRUE)

  # ----- Create Directory -----------------------------------------------------

  if ( !dir.exists(logDir) ) {
    if ( createDir ) {
      dir.create(logDir, showWarnings = FALSE, recursive = TRUE)
    } else {
      stop(sprintf("'logDir = %s' is not found and 'createDir = FALSE'", logDir ))
    }
  }

  # ----- Copy old log files ---------------------------------------------------

  try({
    # NOTE:  Intentionally create timestamp in UTC
    timestamp <- strftime(lubridate::now(tzone = "UTC"), "%Y-%m-%dT%H:%M:%SZ")
    for (logLevel in c("TRACE", "DEBUG", "INFO", "WARN", "ERROR")) {
      oldFile <- file.path(logDir, sprintf("%s%s.log", filePrefix, logLevel))
      newFile <- file.path(logDir, sprintf("%s%s.log.%s", filePrefix, logLevel, timestamp))
      if ( file.exists(oldFile) ) {
        file.rename(oldFile, newFile)
      }
    }
  }, silent = TRUE) %>%
  stopOnError("could not rename old log files")

  # ----- Set up logging -------------------------------------------------------

  try({
    logger.setup(
      traceLog = file.path(logDir, sprintf("%sTRACE.log", filePrefix)),
      debugLog = file.path(logDir, sprintf("%sDEBUG.log", filePrefix)),
      infoLog = file.path(logDir, sprintf("%sINFO.log", filePrefix)),
      warnLog = file.path(logDir, sprintf("%sWARN.log", filePrefix)),
      errorLog = file.path(logDir, sprintf("%sERROR.log", filePrefix))
    )
  }, silent = TRUE) %>%
  stopOnError("could not create log files")

  invisible(NULL)
}

#' Set a variable to a default value if it is NULL
#'
#' Returns `default` when `target` is `NULL`; otherwise returns `target`
#' unchanged.
#'
#' This is useful for assigning default values to optional arguments while
#' preserving any user-supplied value exactly as provided.
#'
#' Optionally, `enforcedType` may be used to coerce the returned value to a
#' specific type. This coercion is applied after the `NULL` check and affects
#' both `target` and `default`.
#'
#' @param target Object to test for `NULL`.
#' @param default Object to return when `target` is `NULL`.
#' @param enforcedType Optional character string specifying the suffix of an
#'   `as.*()` coercion function to apply to the returned value. For example,
#'   `"double"` uses `as.double()`, `"character"` uses `as.character()`,
#'   and `"Date"` uses `as.Date()`.
#'
#'   If `NULL` (the default), no coercion is performed.
#'
#' @return
#' The value of `target` if it is not `NULL`; otherwise `default`.
#'
#' If `enforcedType` is specified, the returned value is coerced using the
#' corresponding `as.*()` function.
#'
#' @export
#'
#' @examples
#' setIfNull(NULL, "foo")
#' setIfNull(10, 0)
#' setIfNull("15", 0)
#'
#' # User-supplied values are returned unchanged
#' setIfNull("15", 0)
#' setIfNull("mean", 0)
#' setIfNull(mean, 0)
#'
#' # Optional type enforcement
#' setIfNull("15", 0, enforcedType = "double")
#' setIfNull(NULL, "15", enforcedType = "integer")
#'
setIfNull <- function(
    target,
    default,
    enforcedType = NULL
) {

  # Return target or default --------------------------------------------------

  result <- if (is.null(target)) default else target


  # Optionally enforce type ---------------------------------------------------

  if (!is.null(enforcedType)) {

    conversionFuncName <- paste0("as.", enforcedType)

    if (!exists(conversionFuncName, mode = "function")) {
      stop(
        sprintf(
          "No coercion function '%s()' found.",
          conversionFuncName
        )
      )
    }

    conversionFunc <-
      get(conversionFuncName, mode = "function")

    result <- conversionFunc(result)

  }

  # Return result -------------------------------------------------------------

  return(result)

}

test_that("setIfNull() returns default only when target is NULL", {

  expect_identical(setIfNull(NULL, "foo"), "foo")
  expect_identical(setIfNull(NULL, 10), 10)
  expect_identical(setIfNull(NULL, TRUE), TRUE)

})

test_that("setIfNull() returns supplied target unchanged", {

  expect_identical(setIfNull("15", 0), "15")
  expect_identical(setIfNull(10, "foo"), 10)
  expect_identical(setIfNull(TRUE, 0), TRUE)
  expect_identical(setIfNull(mean, 0), mean)
  expect_identical(setIfNull(c(1, 2, 3), 0), c(1, 2, 3))

})

test_that("setIfNull() optionally enforces output type with as.* functions", {

  expect_identical(
    setIfNull("15", 0, enforcedType = "double"),
    15
  )

  expect_identical(
    setIfNull(NULL, "15", enforcedType = "integer"),
    15L
  )

  expect_identical(
    setIfNull(1, 0, enforcedType = "character"),
    "1"
  )

  expect_identical(
    setIfNull("2025-01-01", NULL, enforcedType = "Date"),
    as.Date("2025-01-01")
  )

})

test_that("setIfNull() errors when requested coercion function does not exist", {

  expect_error(
    setIfNull(1, 0, enforcedType = "notAType"),
    "No coercion function 'as.notAType\\(\\)' found"
  )

})

#' Manage cache size
#'
#' Remove old or excess files from a cache directory.
#'
#' Files are eligible for removal when their extension matches `extensions`.
#' Matching is case-sensitive and extensions may be supplied with or without a
#' leading dot.
#'
#' Files can be removed for two reasons:
#'
#' \itemize{
#'   \item files older than `maxFileAge` days are removed first
#'   \item if the remaining cache exceeds `maxCacheSize`, additional files are
#'   removed until the cache is under the requested size
#' }
#'
#' When removing files to satisfy `maxCacheSize`, files are ordered by the
#' timestamp specified by `sortBy`.
#'
#' @param cacheDir Location of cache directory.
#' @param extensions Vector of file extensions eligible for removal.
#' @param maxCacheSize Maximum cache size in megabytes.
#' @param sortBy Timestamp used to order files for size-based removal. One of
#'   `"atime"`, `"ctime"`, or `"mtime"`.
#' @param maxFileAge Maximum file age in days. Files with modification times
#'   older than this value are removed regardless of cache size. Fractional days
#'   are allowed.
#'
#' @return
#' Invisibly returns the number of files removed.
#'
#' @details
#' Timestamp meanings are:
#'
#' \describe{
#'   \item{`atime`}{File access time, updated when a file is opened.}
#'   \item{`ctime`}{File change time, updated when file metadata changes.}
#'   \item{`mtime`}{File modification time, updated when file contents change.}
#' }
#'
#' @examples
#' CACHE_DIR <- tempdir()
#'
#' write.csv(matrix(1, 400, 500), file = file.path(CACHE_DIR, "m1.csv"))
#' write.csv(matrix(2, 400, 500), file = file.path(CACHE_DIR, "m2.csv"))
#' write.csv(matrix(3, 400, 500), file = file.path(CACHE_DIR, "m3.csv"))
#' write.csv(matrix(4, 400, 500), file = file.path(CACHE_DIR, "m4.csv"))
#'
#' for (file in list.files(CACHE_DIR, pattern = "\\.csv$", full.names = TRUE)) {
#'   print(file.info(file)[, c("size", "mtime")])
#' }
#'
#' # Remove files based on access time until the cache is under 1 MB
#' manageCache(
#'   CACHE_DIR,
#'   extensions = "csv",
#'   maxCacheSize = 1,
#'   sortBy = "atime"
#' )
#'
#' for (file in list.files(CACHE_DIR, pattern = "\\.csv$", full.names = TRUE)) {
#'   print(file.info(file)[, c("size", "mtime")])
#' }
#'
#' @name manageCache
#' @importFrom rlang .data
#' @export
#'

manageCache <- function(
  cacheDir = NULL,
  extensions = c('html','json','pdf','png'),
  maxCacheSize = 100,
  sortBy = 'atime',
  maxFileAge = NULL
) {

  # Validate parameters --------------------------------------------------------

  stopIfNull(cacheDir)
  stopIfNull(extensions)
  stopIfNull(maxCacheSize)
  stopIfNull(sortBy)

  # Get file info --------------------------------------------------------------

  # Convert incoming size from megabytes to bytes
  maxCacheSize <- as.numeric(maxCacheSize) * 1e6

  # Get all files appropriate for deletion
  filesList <- list()
  for ( extension in extensions ) {
    extension <- stringr::str_replace(extension,'^\\.','') # replace initial '.'
    pattern <- paste0('\\.',extension,'$')
    filesList[[extension]] <- list.files(cacheDir,
                                         pattern=pattern,
                                         full.names=TRUE)
  }
  cacheFiles <- unlist(filesList, use.names=FALSE)

  # Create a dataframe with access times and file sizes
  cacheDF <- file.info(cacheFiles)
  cacheDF$file <- rownames(cacheDF)

  # Remove old files -----------------------------------------------------------

  if ( is.null(maxFileAge) ) {
    ageRemovalCount <- 0
  } else {
    expiration <- lubridate::now(tzone = "UTC") - lubridate::ddays(maxFileAge)
    removalDF <- dplyr::filter(cacheDF, cacheDF$mtime < expiration)
    ageRemovalCount <- nrow(removalDF)
    if ( ageRemovalCount > 0 ) {
      file.remove(removalDF$file)
      # Remove deleted files before size-based cleanup.
      cacheDF <-
        cacheDF %>%
        dplyr::filter(!.data$file %in% removalDF$file)
    }
  }

  # Remove excess files --------------------------------------------------------

  # Use dplyr to order by value specified by sortBy
  if ( !sortBy %in% c("atime", "ctime", "mtime") ) {
    stop("invalid value for parameter 'sortBy'")
  } else {
    sizeByDF <- dplyr::arrange(cacheDF, dplyr::desc(.data[[sortBy]]))
  }

  # Compute a running total
  sizeByDF$cumulativeSize <- cumsum(sizeByDF$size)

  # Remove all files associated with cumulativeSize > maxCacheSize
  removalMask <- sizeByDF$cumulativeSize > maxCacheSize
  removalFiles <- sizeByDF$file[removalMask]
  sizeRemovalCount <- length(removalFiles)
  if ( sizeRemovalCount > 0 ) {
    file.remove(removalFiles)
  }

  # Return ---------------------------------------------------------------------

  removalCount <- ageRemovalCount + sizeRemovalCount

  return(invisible(removalCount))

}

test_that("manageCache() tests sortBy values", {

  oneByte <- 1e-6

  expect_error({
    removedCount <- manageCache(
      tempdir(),
      extensions = ".MazamaCoreUtils-test",
      maxCacheSize = oneByte,
      sortBy = "a"
    )
  }, NULL) # expects error

  expect_error({
    removedCount <- manageCache(
      tempdir(),
      extensions = ".MazamaCoreUtils-test",
      maxCacheSize = oneByte,
      sortBy = "atime"
    )
  }, NA) # expects no error

})

test_that("manageCache() doesn't remove files when maxCacheSize is big", {

  # setup
  oneTByte <- 1e6
  count <- 4

  for ( i in 1:count ) {
    write.csv(iris, tempfile(fileext = ".MazamaCoreUtils-test"))
  }

  removedCount <- manageCache(
    tempdir(),
    extensions = ".MazamaCoreUtils-test",
    maxCacheSize = oneTByte
  )

  expect_equal(removedCount, 0)

  # cleanup
  file.remove(list.files(
    tempdir(),
    pattern = ".MazamaCoreUtils-test",
    full.names = TRUE
  ))

})

test_that("manageCache() removes files when maxCacheSize is small", {

  # setup
  count <- 4
  oneByte <- 1e-6

  for ( i in 1:count ) {
    write.csv(iris, tempfile(fileext = ".MazamaCoreUtils-test"))
  }

  removedCount <- manageCache(
    tempdir(),
    extensions = ".MazamaCoreUtils-test",
    maxCacheSize = oneByte
  )

  expect_equal(removedCount, count)
})

---
title: "Logging"
author: "Jonathan Callahan"
date: "2018-12-02"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Logging}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Logging in Java and Python

Operational systems, by definition, need to work without human input. Systems
are considered "operational" after they have been thoroughly tested and shown to
work properly with a variety of input.

However, no software is perfect and no real-world system operates with 100% 
availability or 100% consistent input. Things occasionally go wrong -- perhaps
intermittently. In a situation with occasional failures it is vitally important
to have good logging information for forensic analysis.

Other languages used in operational settings support logging at different levels
of information density. The code to write out logs at these different levels
looks very similar in java and python:

**java**
```
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.impl.Jdk14Logger;
```
... ugly setup of log files ...
```
log.error("ERROR level message");
log.warn("WARN level message");
log.info("INFO level message");
log.debug("DEBUG level message");
log.trace("TRACE level message");
```

**python**
```
import logging
 
logging.basicConfig(filename="my_INFO.log",
                    level=logging.INFO)

logging.error("ERROR level message")
logging.warning("WARNING level message")
logging.info("INFO level message")
logging.debug("DEBUG level message")
logging.trace("TRACE level message")
```

# Logging in R

The **MazamaCoreUtils** package provides functions so that R logging can look
very similar:

```
library(MazamaCoreUtils)
 
logger.setup(infoLog="my_INFO.log")

logger.error("ERROR level message")
logger.warn("WARN level message")
logger.info("INFO level message")
logger.debug("DEBUG level message")
logger.trace("TRACE level message")
```

This functionality is built on top of the excellent
[logger](https://daroczig.github.io/logger/) package.

## Log Levels

When used properly, logging at different levels lets you create log files 
that help you quickly navigate to the specific part of your code base that 
generated an error.

Different log levels also allow problems to be dealt with by staff with different
skill levels. Core developers don't need to be dragged in whenever anything goes
wrong. If a real-time web site is non-responsive, non-technical staff can be 
assigned to look at logs to assess the situation.

 * If the `ERROR` or `WARN` logs report that no data was obtained from an upstream 
source, the problem is not with the R code.
 * If the `INFO` log reports that incoming data was properly obtained but that
none of it passed QC thresholds, the problem is not with the R code.
 * If, however, these logs show everything running smoothly until some ugly R
error message, then it is time to call in the core developers.

Here are our best practices for how to use the different log levels:

### ERROR

`ERROR` log statements should only be generated immediately before a `stop()` is
issued. The `ERROR.log` file should only have entries when R code failed to
complete.

### WARN

`WARN` log statements imply that something unexpected happened but that R code is
able to complete, perhaps with unpredictable results. An appropriate use would
be after testing the response of an internet request. R code may be able to
trap and handle a failed request but should comment on the failure in the logs.

### INFO

`INFO` log statements are useful for a general overview of how your R code is
proceeding, reflecting top level code structure and perhaps ending with a final:

```
  logger.info("Successfully completed ...")
  logger.info("\n=============== THE END ===============\n")
```

### DEBUG

`DEBUG` log statements are best used to map out the path taken through the
code. We recommend including `DEBUG` level statements inside of top level loops
and beginning each function with the following:

```
logger.debug("----- FUNCTION_NAME() -----")
```

### TRACE

`TRACE` log statements are for the gory details of any part of your code where
you expect or are experiencing errors. These should contain information about
program state and are intended only for developers.

# A Logging Example

The following minimal example demonstrates the use of logging statements and
their output. 

```{r loggingExample}
library(MazamaCoreUtils)

# Set up logging 
tmpDir <- tempdir()
logger.setup(
  errorLog = file.path(tmpDir,"ERROR.log"),
  infoLog = file.path(tmpDir,"INFO.log"),
  traceLog = file.path(tmpDir,"TRACE.log")
)

# Begin main program
logger.info("Begin loggingExample ...")

# Try something that might fail
outputSums <- list()
for ( column in names(iris) ) {
  
  logger.trace("Working on: %s", column)
  
  result <- try({
    outputSums[[column]] <- 
      paste0("Sum of ", column, " = ", sum(iris[[column]]))
  }, silent = TRUE)
  
  if ( "try-error" %in% class(result) ) {
    logger.warn("Oops: %s", geterrmessage())
  }
  
}

# Successful completion
logger.error("NO ERRORS")
logger.info("Successfully completed loggingExample")
logger.info("\n=============== THE END ===============\n")

# Check the output
str(outputSums)
```

Here is what is generated in the log files:

## ERROR.log

```{r errorLog, echo = TRUE}
cat(paste(readLines(file.path(tmpDir,"ERROR.log")),collapse='\n'))
```

## INFO.log

```{r infoLog, echo = TRUE}
cat(paste(readLines(file.path(tmpDir,"INFO.log")),collapse='\n'))
```

## TRACE.log

```{r traceLog, echo = TRUE}
cat(paste(readLines(file.path(tmpDir,"TRACE.log")),collapse='\n'))
```

# Extra Functionality

Two more utility functions need to be mentioned

### `logger.setLevel()`

This function sets the log level used in the console. By default this level is
set to `FATAL` which basically means "only when R crashes". If you are debugging
code interactively you may wish to see log messages as the are generated by
entering the following command:

```
logger.setLevel(TRACE)
```

### `initializeLogging()`

This function is a wrapper for our typical usage. It accepts a `logDir`
parameter and looks for `TRACE`, `DEBUG`, `INFO` and `ERROR` logs in that 
directory. If they are found, they are renamed as 
`paste0(logLevel,".log.",timestamp)`. After renaming, new log files are setup.

This is particularly useful for operational code run inside docker containers
which may be automatically rebooted.

---
title: "Cache Management"
author: "Thomas Bergamaschi"
date: "2018-10-18"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Cache Management}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

When building a web service, it is desirable to save commonly requested products
in a cache directory to avoid time wasted reproducing them unnecessarily.
Because the cache has finite disk space allocated to it, the cache should be 
routinely purged of old or outdated files to make room new ones. The 
```manageCache()``` utility function simplifies this process.

## A Product Cache Example

Lets first make a cache directory and put some data products in it.

```{r createCache}
# Create a cache directory
CACHE_DIR <- file.path(tempdir(), 'cache')
if ( file.exists(CACHE_DIR) == FALSE ) {
  dir.create(CACHE_DIR)
}

# Add a few files to the cache
write.csv(matrix(1,400,500), file=file.path(CACHE_DIR,'m1.csv'))
Sys.sleep(1) # wait a bit between each to give them different mtimes
write.csv(matrix(2,400,500), file=file.path(CACHE_DIR,'m2.csv'))
Sys.sleep(1)
write.csv(matrix(3,400,500), file=file.path(CACHE_DIR,'m3.csv'))
Sys.sleep(1)
write.csv(matrix(4,400,500), file=file.path(CACHE_DIR,'m4.csv'))
```

We can look in our new cache directory and see the four files we just added. 
The directory contains about 1.5 MB of data.

```{r checkCache}
cachedFiles <- list.files(CACHE_DIR, full.names = TRUE)
infoDF <- file.info(cachedFiles)
cacheSize = (sum(infoDF$size) / 1e6) # in MB
print(list.files(CACHE_DIR))
sprintf("Cache size = %s MB", cacheSize)
```

In order to simulate file requests, lets read two of them to update their access 
time.

```{r accessFiles, echo=TRUE}
# Access two of the files, updating their atime
invisible( read.csv(file.path(CACHE_DIR, 'm1.csv')) )
invisible( read.csv(file.path(CACHE_DIR, 'm2.csv')) )
```

Now, lets use ```manageCache()``` to get our cache down to 1 MB.

```{r manageCache}
# Use manageCache() to get cache to 1 MB
library(MazamaCoreUtils)
manageCache(CACHE_DIR, extensions = 'csv', maxCacheSize = 1)
```

When we check our cache again, we will see that the two files with the oldest
access times are gone and the cache size is now under 1 MB.

```{r checkCacheAgain}
# Check cache contents and total size again
cachedFiles <- list.files(CACHE_DIR, full.names = TRUE)
infoDF <- file.info(cachedFiles)
cacheSize = (sum(infoDF$size) / 1e6) # in MB
print(list.files(CACHE_DIR))
sprintf("Cache size = %s MB", cacheSize)
```

## Removing 'Stale' Products

Web services that provide access to real-time data often generate products that
have an expiration date. Files older than a specific number of days or hours 
should be removed from the cache because they no longer represent the current
status. Removing stale files can also help to keep the cache much smaller than 
the absolute maximum cache size, enhancing overall performance.

Stale files -- files that haven't been modified in a while -- can be removed 
regardless of  cache size with the `maxFileAge` parameter. When this is set, 
files with an  `mtime` older than `maxFileAge` will  be removed before any test 
of the  `maxCacheSize`. Fractional days are allowed.

You can remove standard products in the cache that haven't been modified in the 
last 3 hours with:

```{r maxFileAge}
manageCache(CACHE_DIR, maxFileAge = 3/24)
```

## Other Use Cases

When used to manage a product cache, the most typical behavior will be to sort
files based on last access time. The `manageCache()` function uses
`sortBy = "atime"` as the default. It is also possible to sort based on 
modification time `mtime` or change time `ctime`.

The use case scenario for `sortBy = "mtime"` might involve files that are
considered *stale* if the contents aren't updated.

A use case scenario for `sortBy = "ctime"` is not clear.

---
title: "Error Handling"
author: "Jonathan Callahan"
date: "2018-12-02"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Error Handling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Error Handling in Java and Python

Operational systems, by definition, need to work without human input. Systems
are considered "operational" after they have been thoroughly tested and shown to
work properly with a variety of input.

However, no software is perfect and no real-world system operates with 100% 
availability or 100% consistent input. Things occasionally go wrong -- perhaps
intermittently. In a situation with occasional failures it is vitally important
to have robust error handling to deal with both expected and unexpected results.

Other languages used in operational settings have language statements to help
with error handling. The code to handle errors looks very similar in java and 
python:

**java**
```
try {
  myFunc(a)
} catch (abcException e) {
  // handle abcException
} catch (defException e) {
  // handle defException
} finally {
  // always executed after handlers
}
```

**python**
```
try:
  myFunc(a)
except abcError:
  # handle abcError
except defError:
  # handle defError
finally:
  # always executed after handlers
```

# Error Handling in R

Not surprisingly, a functional language like R does things differently:

* No language statements for error handling
* `tryCatch()` is a function
* Need to define warning handler function
* Need to define error handling function
* Scope issues

Nevertheless, R's error handling functions can be made to look similar to java
and python:

```
result <- tryCatch({
  myFunc(a)
}, warning = function(w) {
  # handle all warnings
}, error = function(e) {
  # handle all errors
}, finally = {
  # always executed after handlers
}
```

For more details see:

* [Basic Error Handling with tryCatch()](https://working-with-data.mazamascience.com/2020/10/08/basic-error-handing-in-r-with-trycatch/)
* [Exceptions and debugging - Advanced R](http://adv-r.had.co.nz/Exceptions-Debugging.html)

# Simpler Error Handling in R

In our experience, R's error handling is too complicated for simple use
and requires too much from folks who don't consider themselves R-gurus.

Instead, we recommend wrapping any block of code that needs error handling in
a `try()` function and then testing the result to see if an error occurred.

* `try()` is a wrapper around `tryCatch()`
* `try()` ignores warnings
* `try()` returns a “try-error” object on error
* `geterrmessage()` returns latest error msg

For more details see:

* [Easier Error Handling with try()](https://working-with-data.mazamascience.com/2020/10/09/easier-error-handling-in-r-with-try/)

This strategy makes it easy to create error handling logic and easy to
understand what it does. In the following pseudo-code please note that R 
considers everything between `{}` to be a single *expression*:

```
result <- try({
  # ...
  # lines of R code
  # ...
}, silent = TRUE)

if ( "try-error" %in% class(result) ) {
err_msg <- geterrmessage()
  # logging of error message
  # detection and handling of particular error strings
  # stop() if necessary with user friendly error strings
}
```

# `stopOnError()`

The **`stopOnError()`** utility function regularizes our handling of 
errors in operational code. This function tests the first argument for a class of `try-error`
and, if true, performs the following actions:

 1. creates `err_msg` from a user provided error message or, if NULL, 
`geterrmessage()`
 1. allows modification of error messages with arguments `prefix` and `maxLength`
 1. logs this message with `logger.error(err_msg)` if logging has been enabled
 1. throws an updated error message with `stop(err_msg)`

Encouraging junior R programmers to add error handling to their code is now much
easier. They can place any block of R code within a "try block" with the 
following minimal syntax:

```
result <- try({
  # ...
  # lines of R code
  # ...
}, silent = FALSE)
stopOnError(result)
```

Using the `%>%` pipe operator, we can write this even more concisely without 
creating the interim `result` object:

```
try({
  # ...
  # lines of R code
  # ...
}, silent = FALSE) %>%
  stopOnError()
```

# Working example

Here is a working example demonstrating how a web service might test for user 
input that may not have been converted from character to numeric. All errors are
appropriately logged. _(The outer `try()` blocks in the examples below allow the 
code to be evaluated for this vignette.)_. 

In the third example, we see how low level error messages that may be hard to
understand in the context of a complex, multi-level piece of code can be converted
into a message that makes sense in the context of a web service application.

```{r stopOnError}
library(MazamaCoreUtils)
logger.setup()
logger.setLevel(TRACE) # force logs to be printed to the console

# Arbitrarily deep in the stack we might have:
myFunc <- function(x) {
  return(log(x))
}

# ----- Example 1:  good user input --------------------------------------------
try({
  
  userInput <- 10
  logger.trace("class(userInput) = %s", class(userInput))
  
  try({
    myFunc(x = userInput)
  }, silent = TRUE) %>%
    stopOnError()
  
  logger.trace("Continue processing ...")
  
}, silent = TRUE)

# ----- Example 2:  bad user input ---------------------------------------------
try({
  
  userInput <- "10"
  logger.trace("class(userInput) = %s", class(userInput))
  
  try({
    myFunc(x = userInput)
  }, silent = TRUE) %>%
    stopOnError()
  
  logger.trace("Continue processing ...") # we don't get here
  
}, silent = TRUE)

# ----- Example 3:  bad user input, custom error message -----------------------
try({
  
  try({
    logger.trace("class(userInput) = %s", class(userInput))
    myFunc(x = userInput)
  }, silent = TRUE) %>%
    stopOnError("Unable to process user input")
  
  logger.trace("Continue processing ...") # we don't get here
  
}, silent = TRUE)
```

---
title: "Date Parsing"
author: "Jonathan Callahan"
date: "2019-08-20"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Date Parsing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Environmental Time Series Data

Dates, times and timezones can be frustrating, especially when working with
environmental time series such as those collected by air and water quality
sensors.

Environmental time series data often have a strong diurnal signal and are
typically plotted with a time axis displaying local time. However, when 
data are aggregated into larger collections, it is typical to store data with 
a universal time axis -- UTC.

Problems can arise when parsing and formatting dates and times because R
defaults to the system timezone available with `Sys.timezone()`. Imagine
an agency scientist based in Washington, DC, using their laptop to display 
recent air quality data from Los Angeles while at a conference in Tasmania. The 
data center processing  the data might be in Boulder but the data processing 
machine  might be set to use UTC. Potential timezones 
(available with `OlsonNames()`) relevant to this scenario include:

* `America/New_York`
* `America/Los_Angeles`
* `Australia/Tasmania`
* `America/Denver`
* `UTC`

Which timezone should be used to convert a request for data from "2019-08-08" to 
"2018-08-15" into `POSIXct` datetimes?

To enforce specification of timezones and to help with the common user interface
need to specify a range of dates or times, the **MazamaCoreUtils** package
provides the following functions:

* `dateRange()` -- parses and returns `POSIXct` start and end dates 
representing full days in the specified timezone
* `timeRange()` -- parses and returns `POSixct` start and end times in the 
specified timezone
* `parseDatetime()` -- parses and returns a vector of `POSIXct` values in the
specified timezone

The `parseDatetime()` function is intended as a timezone-requiring replacement 
for `lubridate::parse_date_time()`.

# Linting for timezones

Enforcing the specification of timezones throughout a body of code is the most
robust way to remove timezone-related errors from your software. To help with this
this type of code review, the package also includes functions for testing whether
specific named arguments are used with certain function calls:

* `lintFunctionArgs_file()` -- check a single file
* `lintFunctionArgs_dir()` -- check an entire directory

To use these functions you must define a set of `function:argument` rules to be 
applied such as:

```
timezoneLintRules <- list(
  "parse_date_time" = "tz",
  "with_tz" = "tzone",
  "now" = "tzone",
  "strftime" = "tz"
)
```

This is interpreted as:

* Every use of the `parse_date_time()` function must use the `tz` argument
explicitly.
* Every use of the `with_tz()` function must use the `tzone` argument explicitly
* ...

While these functions could be used to test for explicit use in any 
`function:argument` pair, our concern here is primarily with specification of
timezones. The package includes a detailed list of `timezoneLintRules` to help 
with this. As an example, here is the result of linting the `dateRange.R`
function in this package:

```
> lintFunctionArgs_file("R/dateRange.R", timezoneLintRules)
# A tibble: 7 x 6
  file        line_number column_number function_name   named_args includes_required
  <chr>             <int>         <int> <chr>           <list>     <lgl>            
1 dateRange.R         125            29 with_tz         <chr [1]>  TRUE             
2 dateRange.R         128            27 with_tz         <chr [1]>  TRUE             
3 dateRange.R         141            18 parse_date_time <chr [2]>  TRUE             
4 dateRange.R         142            18 parse_date_time <chr [2]>  TRUE             
5 dateRange.R         159            18 parse_date_time <chr [2]>  TRUE             
6 dateRange.R         176            18 parse_date_time <chr [2]>  TRUE             
7 dateRange.R         188            18 now             <chr [1]>  TRUE             
```

The result shows that the `dateRange.R` source code is consistent in always
explicitly specifying a timezone.

Hopefully, this attention to timezones will help your code avoid 
misunderstandings when it comes to date and time requests.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-location.R
\name{validateLonsLats}
\alias{validateLonsLats}
\title{Validate longitude and latitude vectors}
\usage{
validateLonsLats(longitude = NULL, latitude = NULL, na.rm = FALSE)
}
\arguments{
\item{longitude}{Vector of longitudes in decimal degrees east.}

\item{latitude}{Vector of latitudes in decimal degrees north.}

\item{na.rm}{Logical specifying whether to remove \code{NA} values before
validation.}
}
\value{
Invisibly returns \code{TRUE} if validation succeeds.
}
\description{
Validate longitude and latitude vectors to ensure they are numeric,
have matching lengths, and contain values within valid geographic bounds.
}
\details{
Longitudes must fall between -180 and 180 degrees and latitudes must
fall between -90 and 90 degrees. If validation fails, an error is
generated.
}
\examples{
longitude <- c(-122.5, -122.4)
latitude <- c(47.5, 47.6)

validateLonsLats(longitude, latitude)

# Remove missing values before validation
validateLonsLats(
  c(-122.5, NA),
  c(47.5, NA),
  na.rm = TRUE
)

\dontrun{
validateLonsLats(c(-200, 0), c(45, 46))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createLocationMask.R
\name{createLocationMask}
\alias{createLocationMask}
\title{Create a mask of valid locations}
\usage{
createLocationMask(
  longitude = NULL,
  latitude = NULL,
  lonRange = c(-180, 180),
  latRange = c(-90, 90),
  removeZeroZero = TRUE
)
}
\arguments{
\item{longitude}{Vector of longitudes in decimal degrees east.}

\item{latitude}{Vector of latitudes in decimal degrees north.}

\item{lonRange}{Range of valid longitudes.}

\item{latRange}{Range of valid latitudes.}

\item{removeZeroZero}{Logical specifying whether the coordinate pair
\verb{(0, 0)} should be treated as invalid.}
}
\value{
Logical vector identifying valid locations.
}
\description{
Create a logical mask identifying valid longitude/latitude pairs.
}
\details{
The returned logical vector contains \code{TRUE} for valid locations and \code{FALSE}
for invalid locations. This is useful for filtering data frames to retain
only records with valid geographic coordinates.

Longitude and latitude values are considered valid when they:

\itemize{
\item fall within \code{lonRange} and \code{latRange}
\item are not missing
\item are not located at \verb{(0, 0)} when \code{removeZeroZero = TRUE}
}

The \code{lonRange} and \code{latRange} arguments can be used to restrict valid
locations to a rectangular geographic region.
}
\examples{
createLocationMask(
  longitude = c(-120, NA, -120, -220, -120, 0),
  latitude = c(45, 45, NA, 45, 100, 0)
)

createLocationMask(
  longitude = -120:-90,
  latitude = 20:50,
  lonRange = c(-110, -100),
  latRange = c(30, 40)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_getLinks.R
\name{html_getLinks}
\alias{html_getLinks}
\alias{html_getLinkNames}
\alias{html_getLinkUrls}
\title{Extract links from an HTML page}
\usage{
html_getLinks(url = NULL, relative = TRUE)

html_getLinkNames(url = NULL)

html_getLinkUrls(url = NULL, relative = TRUE)
}
\arguments{
\item{url}{URL or local file path of an HTML page.}

\item{relative}{Logical specifying whether to return relative URLs. If
\code{FALSE}, relative URLs are converted to absolute URLs using \code{url} as the
base.}
}
\value{
A tibble with \code{linkName} and \code{linkUrl} columns.

\code{html_getLinkNames()} returns a character vector of link names.

\code{html_getLinkUrls()} returns a character vector of link URLs.
}
\description{
Parse an HTML page and return all \verb{<a href="...">...</a>} links as a data
frame.
}
\details{
The returned data frame contains the human-readable link text in \code{linkName}
and the \code{href} value in \code{linkUrl}. This is useful for extracting links from
index pages, including web-accessible directories that list downloadable
files.

Wrapper functions \code{\link[=html_getLinkNames]{html_getLinkNames()}} and \code{\link[=html_getLinkUrls]{html_getLinkUrls()}} return the
corresponding columns as character vectors.
}
\examples{
\dontrun{

# If you want to download lots of USCensus shapefiles
url <- "https://www2.census.gov/geo/tiger/GENZ2019/shp/"

browseURL(url)

dataLinks <- html_getLinks(url)

dataLinks <-
  dataLinks \%>\%
  dplyr::filter(stringr::str_detect(linkName, "us_county"))

head(dataLinks, 10)

html_getLinkNames(url)
html_getLinkUrls(url, relative = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-location.R
\name{validateLonLat}
\alias{validateLonLat}
\title{Validate longitude and latitude values}
\usage{
validateLonLat(longitude = NULL, latitude = NULL)
}
\arguments{
\item{longitude}{Single longitude in decimal degrees east.}

\item{latitude}{Single latitude in decimal degrees north.}
}
\value{
Invisibly returns \code{TRUE} if validation succeeds.
}
\description{
Validate a single longitude/latitude pair to ensure both values are
numeric scalars and fall within valid geographic bounds.
}
\details{
Longitudes must fall between -180 and 180 degrees and latitudes must
fall between -90 and 90 degrees. If validation fails, an error is
generated.
}
\examples{
validateLonLat(-122.5, 47.5)

\dontrun{
validateLonLat(-200, 47.5)
validateLonLat(-122.5, NA)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timeRange.R
\name{timeRange}
\alias{timeRange}
\title{Create a POSIXct time range}
\usage{
timeRange(
  starttime = NULL,
  endtime = NULL,
  timezone = NULL,
  unit = "sec",
  ceilingStart = FALSE,
  ceilingEnd = FALSE
)
}
\arguments{
\item{starttime}{Desired start datetime.}

\item{endtime}{Desired end datetime.}

\item{timezone}{Olson timezone used to interpret incoming datetimes.}

\item{unit}{Unit used for rounding. Passed to \code{\link[lubridate:round_date]{lubridate::floor_date()}} or
\code{\link[lubridate:round_date]{lubridate::ceiling_date()}}.}

\item{ceilingStart}{Logical specifying whether to round the start time up
instead of down.}

\item{ceilingEnd}{Logical specifying whether to round the end time up instead
of down.}
}
\value{
Two-element \code{POSIXct} vector ordered from earliest to latest.
}
\description{
Create an ordered two-element \code{POSIXct} time range from start and end
datetime values.
}
\details{
Input values are converted with \code{\link[=parseDatetime]{parseDatetime()}} using the required
\code{timezone} argument. The resulting start and end times are sorted so the
earlier time is always returned first.

By default, both times are rounded down with \code{\link[lubridate:round_date]{lubridate::floor_date()}} using
the requested \code{unit}. Set \code{ceilingStart = TRUE} or \code{ceilingEnd = TRUE} to
round either endpoint up with \code{\link[lubridate:round_date]{lubridate::ceiling_date()}} instead.
}
\section{POSIXct inputs}{

When \code{startdate} or \code{enddate} are already \code{POSIXct} values, they are first
converted to \code{timezone} with \code{\link[lubridate:with_tz]{lubridate::with_tz()}} without changing the
represented instant in time.
}

\examples{
timeRange(
  starttime = "2019-01-08 10:12:15",
  endtime = 20190109102030,
  timezone = "UTC"
)

timeRange(
  starttime = "2019-01-08 10:12:15",
  endtime = "2019-01-09 10:20:30",
  timezone = "UTC",
  unit = "hour"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-logging.R
\name{logger.fatal}
\alias{logger.fatal}
\title{Python-style logging statements}
\usage{
logger.fatal(msg, ...)
}
\arguments{
\item{msg}{Message with optional format strings.}

\item{...}{Additional arguments passed to \code{sprintf()} formatting.}
}
\value{
No return value. Called for side effects.
}
\description{
Emit a \code{FATAL} level log message.
}
\details{
Logging must first be initialized with \code{\link[=logger.setup]{logger.setup()}}.
}
\seealso{
\code{\link[=logger.setup]{logger.setup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_getTables.R
\name{html_getTables}
\alias{html_getTables}
\alias{html_getTable}
\title{Extract tables from an HTML page}
\usage{
html_getTables(url = NULL, header = NA)

html_getTable(url = NULL, header = NA, index = 1)
}
\arguments{
\item{url}{URL or local file path of an HTML page.}

\item{header}{Logical specifying whether the first row should be used as
column names. If \code{NA}, the first row is used only when it contains \verb{<th>}
elements.}

\item{index}{Index identifying which table to return.}
}
\value{
List of data frames, one for each HTML table.

A single data frame containing the requested HTML table.
}
\description{
Parse an HTML page and return all \verb{<table>} elements as a list of data
frames.
}
\details{
The \code{url} argument may be either a remote URL or a local file path. Tables are
parsed with \code{\link[rvest:html_table]{rvest::html_table()}}. To extract a single table, use
\code{\link[=html_getTable]{html_getTable()}}.
}
\examples{
\dontrun{
url <- "https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"

tables <- html_getTables(url)
firstTable <- tables[[1]]

head(firstTable)
nrow(firstTable)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functionArgument_lintRules.R
\docType{data}
\name{timezoneLintRules}
\alias{timezoneLintRules}
\title{Timezone linting rules}
\format{
A named list of function/argument pairs.
}
\usage{
timezoneLintRules
}
\description{
Rules used by \code{\link[=lintFunctionArgs_file]{lintFunctionArgs_file()}} and \code{\link[=lintFunctionArgs_dir]{lintFunctionArgs_dir()}} to find
date/time function calls that should explicitly specify timezone arguments.
}
\details{
Each list name is a function to check. Each value is the required named
timezone-related argument for that function.

Entries with \code{"DEPRECATED"} are used to flag functions that should generally
be avoided in package code because they depend on the local system clock or
timezone.
}
\examples{
str(timezoneLintRules)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functionArgument_linting.R
\name{lintFunctionArgs}
\alias{lintFunctionArgs}
\alias{lintFunctionArgs_file}
\alias{lintFunctionArgs_dir}
\title{Lint function calls for required named arguments}
\usage{
lintFunctionArgs_file(filePath = NULL, rules = NULL, fullPath = FALSE)

lintFunctionArgs_dir(dirPath = "./R", rules = NULL, fullPath = FALSE)
}
\arguments{
\item{filePath}{Path to a single R source file.}

\item{rules}{Named list of linting rules. Each list name is a function name
and each value is a character vector of required named arguments.}

\item{fullPath}{Logical specifying whether returned file paths should be
absolute paths. If \code{FALSE}, only base file names are returned.}

\item{dirPath}{Path to a directory containing R source files.}
}
\value{
A tibble describing matching function calls, with columns:

\describe{
\item{file}{Source file path or file name.}
\item{line_number}{Line number where the function call begins.}
\item{column_number}{Column number where the function call begins.}
\item{function_name}{Name of the function being checked.}
\item{named_args}{List column containing named arguments used in the call.}
\item{includes_required}{Logical indicating whether all required named
arguments were supplied.}
}
}
\description{
Parse R source code and identify calls to selected functions that are missing
required named arguments.
}
\details{
Rules are supplied as a named list where each name is a function to check and
each value is a character vector of required argument names. A function call
passes when all required arguments are supplied by name.

This linter only checks whether arguments are named in the call. It does not
evaluate code, inspect argument values, or detect unnamed positional
arguments.
}
\section{Limitations}{

This linter only detects named arguments. For example, \code{foo(x = bar, "baz")}
is treated as specifying the named argument \code{x}, but the value \code{bar} and the
unnamed argument \code{"baz"} are not inspected.
}

\examples{
\dontrun{
rules <- list(
  fn_one = "x",
  fn_two = c("foo", "bar")
)

lintFunctionArgs_file(
  filePath = "local_test/timezone_lint_test_script.R",
  rules = rules
)

lintFunctionArgs_dir(
  dirPath = "./R",
  rules = MazamaCoreUtils::timezoneLintRules
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dateSequence.R
\name{dateSequence}
\alias{dateSequence}
\title{Create a POSIXct date sequence}
\usage{
dateSequence(
  startdate = NULL,
  enddate = NULL,
  timezone = NULL,
  ceilingEnd = FALSE
)
}
\arguments{
\item{startdate}{Desired start datetime.}

\item{enddate}{Desired end datetime.}

\item{timezone}{Olson timezone used to interpret incoming dates.}

\item{ceilingEnd}{Logical specifying whether to include the end of the final
day.}
}
\value{
A vector of \code{POSIXct} datetimes at local midnight.
}
\description{
Create a sequence of local-midnight \code{POSIXct} datetimes in a specified
timezone.
}
\details{
The returned sequence begins at midnight local time on \code{startdate} and ends
at midnight local time on \code{enddate}, \emph{i.e.} the \emph{beginning} of \code{enddate}.

The \code{ceilingEnd} argument addresses ambiguity in date ranges such as
\code{"August 1-8"}. With \code{ceilingEnd = FALSE} (default), the sequence ends at
the beginning of August 8. With \code{ceilingEnd = TRUE}, the sequence
includes the entirety of August 8, ending at the midnight that begins August 9.

Input dates are parsed with \code{\link[=parseDatetime]{parseDatetime()}} using the specified
\code{timezone}. Any hour-minute-second information is removed after parsing.
}
\note{
This function preserves local clock-time midnight boundaries across daylight
savings transitions. This differs from \code{seq.Date(..., by = "day")}, which
advances by fixed 24-hour intervals and can drift away from midnight local
time during daylight savings changes.
}
\section{POSIXct inputs}{

When \code{startdate} or \code{enddate} are already \code{POSIXct} values, they are first
converted to \code{timezone} with \code{\link[lubridate:with_tz]{lubridate::with_tz()}} without changing the
represented instant in time. They are then floored to local midnight.
}

\examples{
dateSequence(
  "2019-11-01",
  "2019-11-08",
  timezone = "America/Los_Angeles"
)

dateSequence(
  "2019-11-01",
  "2019-11-07",
  timezone = "America/Los_Angeles",
  ceilingEnd = TRUE
)

# Observe daylight savings handling
datetime <- dateSequence(
  "2019-11-01",
  "2019-11-08",
  timezone = "America/Los_Angeles"
)

datetime
lubridate::with_tz(datetime, "UTC")

# POSIXct inputs preserve the represented instant before flooring
jst <- dateSequence(
  20190307,
  20190315,
  timezone = "Asia/Tokyo"
)

jst

dateSequence(
  jst[1],
  jst[7],
  timezone = "UTC"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-logging.R
\docType{data}
\name{logLevels}
\alias{logLevels}
\alias{FATAL}
\alias{ERROR}
\alias{WARN}
\alias{INFO}
\alias{DEBUG}
\alias{TRACE}
\title{Log levels}
\format{
An object of class \code{integer} of length 1.
}
\usage{
FATAL
}
\description{
Logging level constants used by the MazamaCoreUtils logging API.
}
\details{
Available log levels include:

\preformatted{
FATAL
ERROR
WARN
INFO
DEBUG
TRACE
}

These constants are retained for backwards compatibility with the original
MazamaCoreUtils logging system.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaCoreUtils-package.R
\name{showAPIKeys}
\alias{showAPIKeys}
\title{Show API keys}
\usage{
showAPIKeys()
}
\value{
No return value. Called for side effects.
}
\description{
Print all currently set API keys.
}
\seealso{
\code{\link[=getAPIKey]{getAPIKey()}}, \code{\link[=setAPIKey]{setAPIKey()}}
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createLocationID.R
\name{createLocationID}
\alias{createLocationID}
\title{Create one or more unique location IDs}
\usage{
createLocationID(
  longitude = NULL,
  latitude = NULL,
  precision = 10,
  algorithm = c("geohash", "digest"),
  invalidID = as.character(NA)
)
}
\arguments{
\item{longitude}{Vector of longitudes in decimal degrees east.}

\item{latitude}{Vector of latitudes in decimal degrees north.}

\item{precision}{Precision used when encoding geohashes.}

\item{algorithm}{Encoding algorithm to use. Only \code{"geohash"} is currently
supported. \code{"digest"} is deprecated and will generate an error.}

\item{invalidID}{Identifier to use for invalid locations. This can be a
character string or \code{NA}.}
}
\value{
Character vector of location IDs.
}
\description{
Create a location ID for each longitude/latitude pair using a geohash.
}
\details{
Each location ID is unique within a geohash grid cell. The \code{precision}
argument determines the size of the grid cell. At the equator, approximate
grid cell widths are:

\preformatted{
precision   maximum grid cell width
        5   ~ 4.9 km
        6   ~ 1.2 km
        7   ~ 153 m
        8   ~ 38 m
        9   ~ 4.8 m
       10   ~ 1.2 m
}

Invalid locations are assigned the value specified by \code{invalidID}, typically
\code{NA}.
}
\examples{
longitude <- c(-122.5, 0, NA, -122.5, -122.5)
latitude <- c(47.5, 0, 47.5, NA, 47.5)

createLocationID(longitude, latitude)
createLocationID(longitude, latitude, precision = 7)
createLocationID(longitude, latitude, invalidID = "bad")

}
\references{
\url{https://michaelchirico.github.io/geohashTools/index.html}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaCoreUtils-package.R
\docType{data}
\name{APIKeys}
\alias{APIKeys}
\title{API keys for data services}
\format{
A named list of character strings.
}
\description{
Internal session state used to store API keys for web services.
}
\details{
Users can set API keys with \code{\link[=setAPIKey]{setAPIKey()}}. Keys are remembered for the
duration of the R session and can be retrieved with \code{\link[=getAPIKey]{getAPIKey()}}.

This provides a small abstraction layer for dependent packages so that data
access functions can test for and retrieve provider-specific API keys with
generic code.
}
\seealso{
\code{\link[=getAPIKey]{getAPIKey()}}, \code{\link[=setAPIKey]{setAPIKey()}}, \code{\link[=showAPIKeys]{showAPIKeys()}}
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devtools_check.R
\name{packageCheck}
\alias{packageCheck}
\alias{check_slow}
\alias{check}
\alias{check_fast}
\alias{check_faster}
\alias{check_fastest}
\alias{check_slower}
\alias{check_slowest}
\title{Run package checks}
\usage{
check(pkg = ".")

check_fast(pkg = ".")

check_faster(pkg = ".")

check_fastest(pkg = ".")

check_slow(pkg = ".")

check_slower(pkg = ".")

check_slowest(pkg = ".")
}
\arguments{
\item{pkg}{Package location passed to \code{\link[devtools:check]{devtools::check()}}.}
}
\value{
Invisibly returns the result from \code{\link[devtools:check]{devtools::check()}}.
}
\description{
Convenience wrappers around \code{\link[devtools:check]{devtools::check()}} for package checking at
different levels of thoroughness.
}
\details{
These functions make it easy to run quick checks during active development
and more thorough checks before merging or releasing package changes.

The functions are ordered from most thorough to fastest:

\describe{
\item{\code{check_slowest()}}{
Builds the manual, runs \code{donttest} and \code{dontrun} examples, and
uses \code{--use-gct}.
}
\item{\code{check_slower()}}{
Builds the manual and runs \code{donttest} and \code{dontrun} examples.
}
\item{\code{check_slow()}}{
Builds the manual and runs \code{donttest} examples.
}
\item{\code{check()}}{
Standard development check without building the manual or running
\code{donttest} examples.
}
\item{\code{check_fast()}}{
Skips vignette building and ignores vignettes during checking.
}
\item{\code{check_faster()}}{
Skips vignette building, ignores vignettes, and skips examples.
}
\item{\code{check_fastest()}}{
Skips vignette building, ignores vignettes, skips examples, and skips
tests.
}
}
}
\seealso{
\code{\link[devtools:check]{devtools::check()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-logging.R
\name{logger.isInitialized}
\alias{logger.isInitialized}
\title{Check whether logging has been initialized}
\usage{
logger.isInitialized()
}
\value{
Logical scalar indicating whether logging has been initialized.
}
\description{
Determine whether \code{\link[=logger.setup]{logger.setup()}} has already been called.
}
\details{
This function is useful in package code that conditionally emits log
statements only when logging has been configured.
}
\examples{
\dontrun{
logger.isInitialized()

logger.setup()

logger.isInitialized()
}

}
\seealso{
\code{\link[=logger.setup]{logger.setup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initializeLogging.R
\name{initializeLogging}
\alias{initializeLogging}
\title{Initialize standard log files}
\usage{
initializeLogging(logDir = NULL, filePrefix = "", createDir = TRUE)
}
\arguments{
\item{logDir}{Directory in which to write log files.}

\item{filePrefix}{Character string prepended to log file names.}

\item{createDir}{Logical specifying whether to create \code{logDir} if it does
not already exist.}
}
\value{
No return value. Called for side effects.
}
\description{
Create a standard set of MazamaCoreUtils log files.
}
\details{
This convenience function creates or validates a log directory, archives any
existing standard log files by appending a UTC timestamp, and then initializes
logging with \code{\link[=logger.setup]{logger.setup()}}.

Standard log files include:

\preformatted{
TRACE.log
DEBUG.log
INFO.log
WARN.log
ERROR.log
}

When \code{filePrefix} is supplied, it is prepended to each log file name.
}
\seealso{
\code{\link[=logger.setup]{logger.setup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timeStamp.R
\name{timeStamp}
\alias{timeStamp}
\title{Create character timestamps}
\usage{
timeStamp(datetime = NULL, timezone = NULL, unit = "sec", style = "ymdhms")
}
\arguments{
\item{datetime}{Vector of character, integer, or \code{POSIXct} datetimes.}

\item{timezone}{Olson timezone used to interpret incoming datetimes.}

\item{unit}{Temporal precision of the generated timestamp.}

\item{style}{Output timestamp style.}
}
\value{
Character vector of timestamps.
}
\description{
Convert datetimes to compact character timestamps suitable for file names,
identifiers, labels, and other reproducible text output.
}
\details{
Input values are converted with \code{\link[=parseDatetime]{parseDatetime()}} using the required
\code{timezone} argument. When \code{datetime = NULL}, the current UTC time is used
and \code{timezone} defaults to \code{"UTC"}.

The \code{unit} argument controls the precision of the output timestamp. The
\code{style} argument controls the output format.

Supported \code{unit} values are:

\preformatted{
"year"
"month"
"day"
"hour"
"min"
"sec"
"msec"
}

Supported \code{style} values are:

\preformatted{
"ymdhms"   compact calendar time
"ymdThms"  compact calendar time with "T" separator
"julian"   year and Julian day
"clock"    ISO-like clock time
}

For \code{style = "julian"} and \code{unit = "month"}, the timestamp uses the Julian
day associated with the beginning of the month.
}
\section{POSIXct inputs}{

When \code{startdate} or \code{enddate} are already \code{POSIXct} values, they are first
converted to \code{timezone} with \code{\link[lubridate:with_tz]{lubridate::with_tz()}} without changing the
represented instant in time.
}

\examples{
datetime <- parseDatetime("2019-01-08 12:30:15", timezone = "UTC")

timeStamp()
timeStamp(datetime, "UTC", unit = "year")
timeStamp(datetime, "UTC", unit = "month")
timeStamp(datetime, "UTC", unit = "month", style = "julian")
timeStamp(datetime, "UTC", unit = "day")
timeStamp(datetime, "UTC", unit = "day", style = "julian")
timeStamp(datetime, "UTC", unit = "hour")
timeStamp(datetime, "UTC", unit = "min")
timeStamp(datetime, "UTC", unit = "sec")
timeStamp(datetime, "UTC", unit = "sec", style = "ymdThms")
timeStamp(datetime, "UTC", unit = "sec", style = "julian")
timeStamp(datetime, "UTC", unit = "sec", style = "clock")
timeStamp(datetime, "America/Los_Angeles", unit = "sec", style = "clock")
timeStamp(datetime, "America/Los_Angeles", unit = "msec", style = "clock")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaCoreUtils-package.R
\name{getAPIKey}
\alias{getAPIKey}
\title{Get API key}
\usage{
getAPIKey(provider = NULL)
}
\arguments{
\item{provider}{Web service provider.}
}
\value{
API key string, \code{NULL}, or a named list of all provider/key pairs.
}
\description{
Return the API key associated with a web service provider.
}
\details{
If \code{provider = NULL}, all currently stored API keys are returned.
}
\seealso{
\link{APIKeys}, \code{\link[=setAPIKey]{setAPIKey()}}, \code{\link[=showAPIKeys]{showAPIKeys()}}
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaCoreUtils-package.R
\name{setAPIKey}
\alias{setAPIKey}
\title{Set API key}
\usage{
setAPIKey(provider = NULL, key = NULL)
}
\arguments{
\item{provider}{Web service provider.}

\item{key}{API key.}
}
\value{
Invisibly returns the previous value of the API key.
}
\description{
Set the API key associated with a web service provider.
}
\details{
API keys are stored in package session state and are remembered only for the
duration of the current R session.
}
\seealso{
\code{\link[=getAPIKey]{getAPIKey()}}, \code{\link[=showAPIKeys]{showAPIKeys()}}
}
\keyword{environment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stopIfNull.R
\name{stopIfNull}
\alias{stopIfNull}
\title{Stop if an object is \code{NULL}}
\usage{
stopIfNull(target, msg = NULL)
}
\arguments{
\item{target}{Object to test.}

\item{msg}{Optional error message to display if \code{target} is \code{NULL}.
Must be a character string of length one.}
}
\value{
Invisibly returns \code{target} when it is not \code{NULL}.
}
\description{
Convenience function for validating that an object is not \code{NULL}.
}
\details{
If \code{target} is not \code{NULL}, it is returned invisibly. If \code{target} is
\code{NULL}, the function stops with either a default or user-supplied
error message.

This function is especially useful for validating required function
arguments or for guarding intermediate results in pipelines.
}
\examples{
# Return input invisibly if not NULL
x <- stopIfNull(5)
print(x)

# Useful in pipelines
y <- 1:10
y_mean <-
  y \%>\%
  stopIfNull() \%>\%
  mean()

\dontrun{
# Trigger the default error message
testVar <- NULL
stopIfNull(testVar)

# Trigger a custom error message
stopIfNull(testVar, msg = "This is NULL")

# Make a failing pipeline
z <- NULL
z_mean <-
  z \%>\%
  stopIfNull("This has failed.") \%>\%
  mean()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadDataFile.R
\name{loadDataFile}
\alias{loadDataFile}
\title{Load R data from a URL or local file}
\usage{
loadDataFile(
  filename = NULL,
  dataUrl = NULL,
  dataDir = NULL,
  priority = c("dataDir", "dataUrl")
)
}
\arguments{
\item{filename}{Name of the \code{.rda} file to load.}

\item{dataUrl}{Remote URL directory containing data files.}

\item{dataDir}{Local directory containing data files.}

\item{priority}{First data source to try when both \code{dataDir} and \code{dataUrl}
are supplied.}
}
\value{
Object loaded from the \code{.rda} file.
}
\description{
Load a pre-generated R binary data file from either a local directory or a
remote URL.
}
\details{
This function is intended for use by package-level \verb{*_load()} helper
functions. It allows locally cached data files to be used when available,
avoiding unnecessary internet access.

If both \code{dataDir} and \code{dataUrl} are provided, \code{priority} determines which
source is tried first. If loading from the first source fails, the other
source is used as a fallback.
}
\examples{
\dontrun{
filename <- "USCensusStates_02.rda"
dataDir <- "~/Data/Spatial"
dataUrl <- "http://data.mazamascience.com/MazamaSpatialUtils/Spatial_0.8"

# Load local file
USCensusStates <- loadDataFile(filename, dataDir = dataDir)

# Load remote file
USCensusStates <- loadDataFile(filename, dataUrl = dataUrl)

# Load local file with remote file as backup
USCensusStates <- loadDataFile(
  filename,
  dataDir = dataDir,
  dataUrl = dataUrl,
  priority = "dataDir"
)

# Load remote file with local file as backup
USCensusStates <- loadDataFile(
  filename,
  dataDir = dataDir,
  dataUrl = dataUrl,
  priority = "dataUrl"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parseDatetime.R
\name{parseDatetime}
\alias{parseDatetime}
\title{Parse datetimes}
\usage{
parseDatetime(
  datetime = NULL,
  timezone = NULL,
  expectAll = FALSE,
  isJulian = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{datetime}{Vector of character, numeric, integer, or \code{POSIXct}
datetimes.}

\item{timezone}{Olson timezone used to interpret incoming datetimes.}

\item{expectAll}{Logical value specifying whether to stop if any non-missing
input values fail to parse.}

\item{isJulian}{Logical value specifying whether \code{datetime} should be
interpreted as a Julian date using day-of-year notation.}

\item{quiet}{Logical value passed to \code{\link[lubridate:parse_date_time]{lubridate::parse_date_time()}} to
suppress parsing warnings.}
}
\value{
A \code{POSIXct} vector.
}
\description{
Convert character, numeric, integer, or \code{POSIXct} datetimes to \code{POSIXct}.
}
\details{
This function accepts a variety of compact date/time formats commonly used in
Mazama Science packages, including \code{Y}, \code{Ym}, \code{Ymd}, \code{YmdH}, \code{YmdHM}, and
\code{YmdHMS}. Inputs may be mixed within the same vector.

Examples of equivalent inputs include:

\preformatted{
20181012130900
"2018-10-12-13-09-00"
"2018 Oct. 12 13:09:00"
}

All incoming datetimes are interpreted in the specified \code{timezone}. If
\code{datetime} is already \code{POSIXct}, it is converted to the requested timezone
with \code{\link[lubridate:with_tz]{lubridate::with_tz()}}.

If a character datetime includes signed offset information, such as
\code{"-07:00"}, that offset is used by \code{\link[lubridate:parse_date_time]{lubridate::parse_date_time()}} when
determining the equivalent instant.
}
\section{Mazama Science conventions}{

Within Mazama Science packages, datetimes not already in \code{POSIXct} format are
often represented as compact decimal values with no separators, such as
\code{20181012} or \code{20181012130900}, either as numbers or strings.
}

\section{Implementation}{

\code{parseDatetime()} is a wrapper around \code{\link[lubridate:parse_date_time]{lubridate::parse_date_time()}} that
defines the datetime formats supported by MazamaCoreUtils.
}

\examples{
# All Y[mdHMS] formats are accepted
parseDatetime(2018, timezone = "America/Los_Angeles")
parseDatetime(201808, timezone = "America/Los_Angeles")
parseDatetime(20180807, timezone = "America/Los_Angeles")
parseDatetime(2018080718, timezone = "America/Los_Angeles")
parseDatetime(201808071812, timezone = "America/Los_Angeles")
parseDatetime(20180807181215, timezone = "America/Los_Angeles")

parseDatetime("2018-08-07 18:12:15", timezone = "America/Los_Angeles")
parseDatetime("2018-08-07 18:12:15-07:00", timezone = "UTC")

# Julian days are accepted
parseDatetime(
  2018219181215,
  timezone = "America/Los_Angeles",
  isJulian = TRUE
)

# Mixed vector inputs are accepted
parseDatetime(
  c("2018-10-24 12:00", "201810311200", "2018-11-07 12:00"),
  timezone = "America/New_York"
)

badInput <- c("20181013", NA, "20181015", "181016", "10172018")

# Return NA for dates that cannot be parsed
parseDatetime(badInput, timezone = "UTC", expectAll = FALSE)

\dontrun{
# Fail if any non-missing dates cannot be parsed
parseDatetime(badInput, timezone = "UTC", expectAll = TRUE)
}

}
\seealso{
\code{\link[lubridate:parse_date_time]{lubridate::parse_date_time()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-logging.R
\name{logger.setLevel}
\alias{logger.setLevel}
\title{Set console log level}
\usage{
logger.setLevel(level)
}
\arguments{
\item{level}{Logging threshold level.}
}
\value{
No return value. Called for side effects.
}
\description{
Set the minimum log level displayed in the console.
}
\details{
By default, only \code{FATAL} messages are displayed in the console. This
function allows users to display additional log messages interactively.

Available log levels are:

\preformatted{
TRACE
DEBUG
INFO
WARN
ERROR
FATAL
}
}
\note{
All functionality is implemented with the excellent \pkg{logger} package.
}
\examples{
\dontrun{
# Enable console logging
logger.setup()

# Show DEBUG and higher messages in the console
logger.setLevel(DEBUG)
}

}
\seealso{
\code{\link[=logger.setup]{logger.setup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errorHandling.R
\name{stopOnError}
\alias{stopOnError}
\title{Stop on try-error}
\usage{
stopOnError(
  result,
  err_msg = "",
  prefix = "",
  maxLength = 500,
  truncatedLength = 120,
  call. = FALSE
)
}
\arguments{
\item{result}{Return value from a \code{try()} block.}

\item{err_msg}{Optional custom error message.}

\item{prefix}{Optional text to prepend to the error message.}

\item{maxLength}{Maximum allowed error message length before truncation.}

\item{truncatedLength}{Length of the truncated error message.}

\item{call.}{Logical indicating whether the call should be included in the
error message. Passed to \code{\link[=stop]{stop()}}.}
}
\value{
Returns \code{NULL} if \code{result} is not a \code{"try-error"}; otherwise stops with an
error.
}
\description{
Generate a consistent error message from the result of a \code{try()} block.
}
\details{
This function is intended for production code where potentially fragile
operations are wrapped in \code{try(..., silent = TRUE)}. If \code{result} inherits
from \code{"try-error"}, a cleaned and optionally customized error message is
generated and passed to \code{\link[=stop]{stop()}}.

If \code{result} is not a \code{"try-error"}, the function returns \code{NULL}.
}
\note{
If logging has been initialized, the final error message is logged with
\code{\link[=logger.error]{logger.error()}} before calling \code{\link[=stop]{stop()}}.
}
\examples{
\dontrun{
myFunc <- function(x) {
  log(x)
}

result <- try({
  myFunc("ten")
}, silent = TRUE)

stopOnError(result)

try({
  myFunc("ten")
}, silent = TRUE) \%>\%
  stopOnError(err_msg = "Unable to process user input")

try({
  myFunc("ten")
}, silent = TRUE) \%>\%
  stopOnError(
    prefix = "USER_INPUT_ERROR",
    maxLength = 40,
    truncatedLength = 32
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MazamaCoreUtils-package.R
\docType{package}
\name{MazamaCoreUtils-package}
\alias{MazamaCoreUtils}
\alias{MazamaCoreUtils-package}
\title{MazamaCoreUtils}
\description{
Utility functions for production R code.
}
\details{
MazamaCoreUtils provides a standardized set of utility functions used by
MazamaScience packages and R-based operational systems.

Functionality includes:

\itemize{
\item Python-style logging
\item simple error messaging
\item cache management
\item API key handling
\item date parsing and formatting
\item longitude/latitude validation
\item unique location ID creation
\item source code linting
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/MazamaScience/MazamaCoreUtils}
  \item Report bugs at \url{https://github.com/MazamaScience/MazamaCoreUtils/issues}
}

}
\author{
\strong{Maintainer}: Jonathan Callahan \email{jonathan.s.callahan@gmail.com}

Other contributors:
\itemize{
  \item Eli Grosman \email{eligrosman1@gmail.com} [contributor]
  \item Spencer Pease \email{spencerpease618@gmail.com} [contributor]
  \item Thomas Bergamaschi \email{tbergama@gmail.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dateRange.R
\name{dateRange}
\alias{dateRange}
\title{Create a POSIXct date range}
\usage{
dateRange(
  startdate = NULL,
  enddate = NULL,
  timezone = NULL,
  unit = "sec",
  ceilingStart = FALSE,
  ceilingEnd = FALSE,
  days = 7
)
}
\arguments{
\item{startdate}{Desired start datetime.}

\item{enddate}{Desired end datetime.}

\item{timezone}{Olson timezone used to interpret incoming dates.}

\item{unit}{Temporal precision used for the returned end-of-range value.
One of \code{"day"}, \code{"hour"}, \code{"min"}, or \code{"sec"}.}

\item{ceilingStart}{Logical specifying whether to round \code{startdate} up to the
next day boundary instead of down.}

\item{ceilingEnd}{Logical specifying whether to include the entirety of the
final day.}

\item{days}{Number of days to include when either \code{startdate} or \code{enddate}
is omitted.}
}
\value{
Two-element \code{POSIXct} vector ordered from earliest to latest.
}
\description{
Create a two-element \code{POSIXct} vector representing a date/time range in a
specified timezone.
}
\details{
The returned range is ordered from earliest to latest. The first element
represents the beginning of the requested date range and the second element
represents the end of the requested date range at the requested temporal
precision.

By default, the returned end time is one \code{unit} \emph{before} the beginning of
\code{enddate}. For example:

\preformatted{
dateRange(20190101, 20190102, timezone = "UTC")
[1] "2019-01-01 00:00:00 UTC"
[2] "2019-01-01 23:59:59 UTC"
}

Setting \code{ceilingEnd = TRUE} includes the entirety of \code{enddate}:

\preformatted{
dateRange(
  20190101,
  20190101,
  timezone = "UTC",
  ceilingEnd = TRUE
)
[1] "2019-01-01 00:00:00 UTC"
[2] "2019-01-01 23:59:59 UTC"
}

The \code{ceilingEnd} argument addresses ambiguity in phrases such as
\code{"August 1-8"}. With \code{ceilingEnd = FALSE} (default), the range extends
through the end of August 7, stopping at the midnight boundary where August 8
begins. With \code{ceilingEnd = TRUE}, the range
extends through the end of August 8.

Input dates are parsed with \code{\link[=parseDatetime]{parseDatetime()}} using the specified
\code{timezone}.
}
\section{Default arguments}{

If either \code{startdate} or \code{enddate} is missing, the missing boundary is
calculated using \code{days}.

If both are missing, \code{enddate} defaults to the current day in \code{timezone}
and \code{startdate} is calculated as \code{enddate - days}.
}

\section{End-of-day units}{

The returned end time is adjusted to the last representable value within the
requested unit:

\describe{
\item{\code{unit = "day"}}{End time is midnight at the start of the final day.}
\item{\code{unit = "hour"}}{End time is \code{23:00:00}.}
\item{\code{unit = "min"}}{End time is \code{23:59:00}.}
\item{\code{unit = "sec"}}{End time is \code{23:59:59}.}
}
}

\section{POSIXct inputs}{

When \code{startdate} or \code{enddate} are already \code{POSIXct} values, they are first
converted to \code{timezone} with \code{\link[lubridate:with_tz]{lubridate::with_tz()}} without changing the
represented instant in time.
}

\section{Parameter precedence}{

When parameters conflict, the following rules apply:

\enumerate{
\item If both \code{startdate} and \code{enddate} are supplied, \code{days} is ignored.
\item If \code{startdate} is missing, \code{ceilingStart} is ignored.
\item If \code{enddate} is missing, \code{ceilingEnd} is ignored.
}
}

\examples{
dateRange("2019-01-08", timezone = "UTC")

dateRange("2019-01-08", unit = "min", timezone = "UTC")

dateRange("2019-01-08", unit = "hour", timezone = "UTC")

dateRange("2019-01-08", unit = "day", timezone = "UTC")

dateRange("2019-01-08", "2019-01-11", timezone = "UTC")

dateRange(
  enddate = 20190112,
  days = 3,
  unit = "day",
  timezone = "America/Los_Angeles"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cacheManagement.R
\name{manageCache}
\alias{manageCache}
\title{Manage cache size}
\usage{
manageCache(
  cacheDir = NULL,
  extensions = c("html", "json", "pdf", "png"),
  maxCacheSize = 100,
  sortBy = "atime",
  maxFileAge = NULL
)
}
\arguments{
\item{cacheDir}{Location of cache directory.}

\item{extensions}{Vector of file extensions eligible for removal.}

\item{maxCacheSize}{Maximum cache size in megabytes.}

\item{sortBy}{Timestamp used to order files for size-based removal. One of
\code{"atime"}, \code{"ctime"}, or \code{"mtime"}.}

\item{maxFileAge}{Maximum file age in days. Files with modification times
older than this value are removed regardless of cache size. Fractional days
are allowed.}
}
\value{
Invisibly returns the number of files removed.
}
\description{
Remove old or excess files from a cache directory.
}
\details{
Files are eligible for removal when their extension matches \code{extensions}.
Matching is case-sensitive and extensions may be supplied with or without a
leading dot.

Files can be removed for two reasons:

\itemize{
\item files older than \code{maxFileAge} days are removed first
\item if the remaining cache exceeds \code{maxCacheSize}, additional files are
removed until the cache is under the requested size
}

When removing files to satisfy \code{maxCacheSize}, files are ordered by the
timestamp specified by \code{sortBy}.

Timestamp meanings are:

\describe{
\item{\code{atime}}{File access time, updated when a file is opened.}
\item{\code{ctime}}{File change time, updated when file metadata changes.}
\item{\code{mtime}}{File modification time, updated when file contents change.}
}
}
\examples{
CACHE_DIR <- tempdir()

write.csv(matrix(1, 400, 500), file = file.path(CACHE_DIR, "m1.csv"))
write.csv(matrix(2, 400, 500), file = file.path(CACHE_DIR, "m2.csv"))
write.csv(matrix(3, 400, 500), file = file.path(CACHE_DIR, "m3.csv"))
write.csv(matrix(4, 400, 500), file = file.path(CACHE_DIR, "m4.csv"))

for (file in list.files(CACHE_DIR, pattern = "\\\\.csv$", full.names = TRUE)) {
  print(file.info(file)[, c("size", "mtime")])
}

# Remove files based on access time until the cache is under 1 MB
manageCache(
  CACHE_DIR,
  extensions = "csv",
  maxCacheSize = 1,
  sortBy = "atime"
)

for (file in list.files(CACHE_DIR, pattern = "\\\\.csv$", full.names = TRUE)) {
  print(file.info(file)[, c("size", "mtime")])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-logging.R
\name{logger.setup}
\alias{logger.setup}
\title{Set up Python-style logging}
\usage{
logger.setup(
  traceLog = NULL,
  debugLog = NULL,
  infoLog = NULL,
  warnLog = NULL,
  errorLog = NULL,
  fatalLog = NULL
)
}
\arguments{
\item{traceLog}{File path receiving \code{TRACE} messages.}

\item{debugLog}{File path receiving \code{DEBUG} messages.}

\item{infoLog}{File path receiving \code{INFO} messages.}

\item{warnLog}{File path receiving \code{WARN} messages.}

\item{errorLog}{File path receiving \code{ERROR} messages.}

\item{fatalLog}{File path receiving \code{FATAL} messages.}
}
\value{
No return value. Called for side effects.
}
\description{
Configure level-specific log files using the package logging API.
}
\details{
Logging is built on top of the \href{https://daroczig.github.io/logger/}{logger}
package while retaining the historical MazamaCoreUtils logging interface.

Separate log files can be created for different log levels so that, for
example, an \code{errorLog} contains only \code{ERROR} and \code{FATAL} messages while a
\code{debugLog} contains \code{DEBUG} messages as well as all higher-severity messages.

Any log file argument left as \code{NULL} is disabled and no file will be created
for that level.

After initialization, logging statements can be generated with:
\code{logger.trace()}, \code{logger.debug()}, \code{logger.info()},
\code{logger.warn()}, \code{logger.error()}, and \code{logger.fatal()}.

Log messages are formatted with:

\preformatted{
LEVEL [YYYY-MM-DD HH:MM:SS UTC] message
}

Console logging is enabled by default only for \code{FATAL} messages. Use
\code{\link[=logger.setLevel]{logger.setLevel()}} to display additional log messages in the console.
}
\note{
All functionality is implemented with the excellent \pkg{logger} package.
}
\examples{
\dontrun{
# Create three log files
logger.setup(
  debugLog = "debug.log",
  infoLog = "info.log",
  errorLog = "error.log"
)

# Generate log messages
logger.trace("trace statement #\%d", 1)
logger.debug("debug statement")
logger.info("info statement \%s \%s", "with", "arguments")
logger.warn("warn statement: \%s", "about to try something risky")

result <- try(1 / "a", silent = TRUE)
logger.error("error message: \%s", geterrmessage())
logger.fatal("fatal statement: \%s", "THE END")

cat(readLines("debug.log"), sep = "\n")
cat(readLines("info.log"), sep = "\n")
cat(readLines("error.log"), sep = "\n")
}

}
\seealso{
\code{\link[=logger.trace]{logger.trace()}}, \code{\link[=logger.debug]{logger.debug()}}, \code{\link[=logger.info]{logger.info()}},
\code{\link[=logger.warn]{logger.warn()}}, \code{\link[=logger.error]{logger.error()}}, \code{\link[=logger.fatal]{logger.fatal()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-tidyeval-data.R
\name{tidyeval-data}
\alias{tidyeval-data}
\alias{.data}
\title{Data pronoun for tidy evaluation}
\description{
See \link[rlang:dot-data]{rlang::.data} for details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setIfNull.R
\name{setIfNull}
\alias{setIfNull}
\title{Set a variable to a default value if it is NULL}
\usage{
setIfNull(target, default, enforcedType = NULL)
}
\arguments{
\item{target}{Object to test for \code{NULL}.}

\item{default}{Object to return when \code{target} is \code{NULL}.}

\item{enforcedType}{Optional character string specifying the suffix of an
\verb{as.*()} coercion function to apply to the returned value. For example,
\code{"double"} uses \code{as.double()}, \code{"character"} uses \code{as.character()},
and \code{"Date"} uses \code{as.Date()}.

If \code{NULL} (the default), no coercion is performed.}
}
\value{
The value of \code{target} if it is not \code{NULL}; otherwise \code{default}.

If \code{enforcedType} is specified, the returned value is coerced using the
corresponding \verb{as.*()} function.
}
\description{
Returns \code{default} when \code{target} is \code{NULL}; otherwise returns \code{target}
unchanged.
}
\details{
This is useful for assigning default values to optional arguments while
preserving any user-supplied value exactly as provided.

Optionally, \code{enforcedType} may be used to coerce the returned value to a
specific type. This coercion is applied after the \code{NULL} check and affects
both \code{target} and \code{default}.
}
\examples{
setIfNull(NULL, "foo")
setIfNull(10, 0)
setIfNull("15", 0)

# User-supplied values are returned unchanged
setIfNull("15", 0)
setIfNull("mean", 0)
setIfNull(mean, 0)

# Optional type enforcement
setIfNull("15", 0, enforcedType = "double")
setIfNull(NULL, "15", enforcedType = "integer")

}
